import gc
import os
import re
import copy

import h5py
import tqdm
import trimesh

try:
    import simnibs
except (ModuleNotFoundError, ImportError):
    print("SimNIBS not found. Some functions might not work.")
import warnings
import collections
import numpy as np

import pynibs


def e_field_gradient_between_wm_gm(
        roi_surf,
        mesh,
        gm_nodes,
        wm_nodes,
        gm_center_distance,
        wm_center_distance
):
    """
    Compute local the E-field gradient at the ROI nodes between the gray and white matter boundary sufaces.
    Adapted from neuronibs/cortical_layer.py/add_e_field_gradient_between_wm_gm_field

    Parameters
    ----------
    roi_surf : simnibs.Msh
        The surface object representing the ROI.
    mesh : simnibs.Msh
        SimNIBS headmesh.
    gm_nodes : np.ndarray, (3, len(roi_surf.nodes))
        For each node of 'roi_surf' (representing, e.g., a cortical layer) the corresponding point on the gray matter surface.
        -> as returned by 'precompute_geo_info_for_layer_field_interpolation'
    wm_nodes : np.ndarray, (3, len(roi_surf.nodes))
        For each node of 'roi_surf' (representing, e.g., a cortical layer) the corresponding point on the white matter surface.
        -> as returned by 'precompute_geo_info_for_layer_field_interpolation'
    gm_center_distance : np.ndarray, (3, len(roi_surf.nodes))
        The distance between the layer nodes and their corresponding point on the gray matter surface.
        -> as returned by 'precompute_geo_info_for_layer_field_interpolation'
    wm_center_distance : np.ndarray, (3, len(roi_surf.nodes))
        The distance between the layer nodes and their corresponding point on the white matter surface.
        -> as returned by 'precompute_geo_info_for_layer_field_interpolation'

    Returns
    -------
    e_field_gradient_per_mm : simnibs.mesh.mesh_io.ElementData
        The E-field gradient from the GM to the WM surface normalized to 1 mm with respect to the gray matter thickness.
        per ROI node
    """
    roi_centers = roi_surf.elements_baricenters().value

    # concatenate the point lists to avoid re-building the (simnibs-internal) data structures
    # for interpolation upon consecutive calls
    to_be_interpolated = np.concatenate((gm_nodes, roi_centers, wm_nodes))
    try:
        interpolated = mesh.field['magnE'].interpolate_scattered(to_be_interpolated)
    except KeyError:
        e_mag = mesh.field['E'].norm()
        interpolated = e_mag.interpolate_scattered(to_be_interpolated)

    num_nodes = gm_nodes.shape[0]
    num_nodes_x2 = num_nodes * 2
    num_nodes_x3 = num_nodes * 3
    e_field_gm = interpolated[:num_nodes]
    e_field_center = interpolated[num_nodes:num_nodes_x2]
    e_field_wm = interpolated[num_nodes_x2:num_nodes_x3]

    e_field_gradient = e_field_gm - e_field_wm
    gray_matter_thickness = gm_center_distance + wm_center_distance
    e_field_gradient_per_mm = e_field_gradient / gray_matter_thickness
    relative_e_field_gradient_per_mm = e_field_gradient_per_mm / e_field_center * 100

    return simnibs.ElementData(relative_e_field_gradient_per_mm, name='rel_gradient_per_mm', mesh=roi_surf)


def e_field_angle_theta(surface, mesh):
    """
     Compute angle between local the E-field vector and surface vector at the ROI nodes.

     Parameters
     ----------
     surface : simnibs.Msh
         The surface object representing the ROI.
     mesh: simnibs.Msh
         The (volumetric) head mesh.

     Returns
     -------
     theta : simnibs.mesh.mesh_io.ElementData
         The angle between local the E-field vector and surface vector for each surface element of the ROI.
     """
    layer_centers = surface.elements_baricenters().value
    interpolated_cell_centers = mesh.field['E'].interpolate_scattered(layer_centers, out_fill='nearest')
    tri_normal = surface.triangle_normals(smooth=30).value
    e_at_roi_cell_centers_normalized = np.divide(
            interpolated_cell_centers,
            np.linalg.norm(interpolated_cell_centers, axis=1)[:, np.newaxis]
    )
    angle = np.array([np.arccos(np.dot(a, b)) / np.pi * 180 for a, b in
                      zip(tri_normal, e_at_roi_cell_centers_normalized)])

    return simnibs.ElementData(angle, name="theta", mesh=surface)


def precompute_geo_info_for_layer_field_interpolation(simnibs_mesh, roi):
    """
    Precomputes geometric properties of the corresponding GM and WM nodes
    in the 'simnibs_mesh' of each node of each layer in ``'roi'``.

    * The corresponding point on the GM and WM surface to each vertex of te ROI surface
      are determined (by raycasting and nearest neighbour search as fallback)
      (interpolation will take place between these nodes)
    * The nodes are moved inside the gray matter by 20 % of their total distance from the GM/WM
      boundary to the midlayer.
    * The distance of the relocated GM and WM nodes to the ROI nodes is determined (required to
      computed the gradient per mm).

    Parameters
    ----------
    simnibs_mesh : simnibs.Msh
        The head model volume mesh.
    roi: pynibs.roi.RegionOfInterestSurface
        The ROI object containing the layers.

    Returns
    -------
    layer_gm_wm_info : dict[str, dict[str,np.ndarray]]
       For each layer defined in the mesh (outer dict key: layer_id),
       provide pre-computed geometrical information as a dictionary (outer dict value):

       * key: gm_nodes
         For each layer node, the corresponding point on the GM surface.
       * key: wm_nodes
         For each layer node, the corresponding point on the WM surface.
       * key: gm_center_distance
         The distance between the layer nodes and the corresponding points on the GM.
       * key: wm_center_distance
         The distance between the layer nodes and the corresponding points on the WM.
    """
    WHITE_MATTER_SURFACE_LABEL = 1001
    GRAY_MATTER_SURFACE_LABEL = 1002

    import vtk

    def make_vtkpolydata(pts, tris):
        # prepare vertices
        pts_vtk = vtk.vtkPoints()
        pts_vtk.SetNumberOfPoints(pts.shape[0])
        for i in range(pts.shape[0]):
            pts_vtk.SetPoint(i, pts[i][0], pts[i][1], pts[i][2])

        # prepare triangles
        tris_vtk = vtk.vtkCellArray()
        for tri in tris:
            tris_vtk.InsertNextCell(3)
            for v in tri:
                tris_vtk.InsertCellPoint(v)

        # prepare GM polygonal surface
        surf_vtk = vtk.vtkPolyData()
        surf_vtk.SetPoints(pts_vtk)
        surf_vtk.SetPolys(tris_vtk)

        return surf_vtk

    gray_matter_surface = simnibs_mesh.crop_mesh(GRAY_MATTER_SURFACE_LABEL)
    white_matter_surface = simnibs_mesh.crop_mesh(WHITE_MATTER_SURFACE_LABEL)

    gm_nodes = gray_matter_surface.nodes.node_coord
    gm_tris = gray_matter_surface.elm.node_number_list
    wm_nodes = white_matter_surface.nodes.node_coord
    wm_tris = white_matter_surface.elm.node_number_list

    # for each point on the layer find corresponding points on the GM/WM surfaces
    # using the VTK implementation of Raycasting
    # make zero indexed and shape (n,3) instead of (n,4), with -1 in the 4th column
    gm_surf_vtk = make_vtkpolydata(gm_nodes, gm_tris[:, :3] - 1)
    wm_surf_vtk = make_vtkpolydata(wm_nodes, wm_tris[:, :3] - 1)

    gm_intersector = vtk.vtkOBBTree()
    gm_intersector.SetDataSet(gm_surf_vtk)
    gm_intersector.BuildLocator()

    wm_intersector = vtk.vtkOBBTree()
    wm_intersector.SetDataSet(wm_surf_vtk)
    wm_intersector.BuildLocator()

    intersectors = [gm_intersector, wm_intersector]
    normal_sign = [1, -1]

    layer_gm_wm_info = dict()
    for layer_idx in range(len(roi.layers)):
        intersec_pts = [[], []]
        layer_id = roi.layers[layer_idx].id
        layer_surf = roi.layers[layer_idx].surface
        layer_centers = layer_surf.elements_baricenters().value
        layer_normals = roi.layers[layer_idx].get_smoothed_normals()

        for pt, normal in zip(layer_centers, layer_normals):
            for idx in range(len(intersectors)):
                intersection_pts = vtk.vtkPoints()
                intersected_tris = vtk.vtkIdList()

                intersector = intersectors[idx]

                intersector.IntersectWithLine(pt, normal_sign[idx] * 100 * normal + pt, intersection_pts,
                                              intersected_tris)
                if intersection_pts.GetNumberOfPoints() > 0:
                    intersec_pts[idx].append(intersection_pts.GetPoint(0))
                else:
                    intersec_pts[idx].append([np.iinfo(np.uint16).max] * 3)

        gm_intersec_pts = np.array(intersec_pts[0])
        wm_intersec_pts = np.array(intersec_pts[1])

        # Check the distances of the found points on the GM/WM surface to the layer nodes.
        # If the distance is too high, raycasting failed (e.g. no intersection found or too far away).
        # In this case, we determine the nearest neighbor for these nodes.
        layer_gm_distance = np.linalg.norm(layer_centers - gm_intersec_pts, ord=2, axis=1)
        layer_wm_distance = np.linalg.norm(layer_centers - wm_intersec_pts, ord=2, axis=1)

        # TODO: consider using a relative distance depending on the local GM thickness
        raycast_error_nodes_gm = np.argwhere(layer_gm_distance > 2)  # mm
        raycast_error_nodes_wm = np.argwhere(layer_wm_distance > 2)  # mm

        closest_gm_nodes, _ = gray_matter_surface.nodes.find_closest_node(layer_centers[raycast_error_nodes_gm],
                                                                          return_index=True)
        closest_wm_nodes, _ = white_matter_surface.nodes.find_closest_node(layer_centers[raycast_error_nodes_wm],
                                                                           return_index=True)

        gm_intersec_pts[raycast_error_nodes_gm] = closest_gm_nodes
        wm_intersec_pts[raycast_error_nodes_wm] = closest_wm_nodes

        layer_gm_distance = np.linalg.norm(layer_centers - gm_intersec_pts, ord=2, axis=1)
        layer_wm_distance = np.linalg.norm(layer_centers - wm_intersec_pts, ord=2, axis=1)

        center_to_gm_vecs = gm_intersec_pts - layer_centers
        center_to_wm_vecs = wm_intersec_pts - layer_centers

        associated_gray_matter_points = gm_intersec_pts - \
                                        np.multiply(
                                                # multiply unit-normals by individual gm thickness, then scale to 20%
                                                center_to_gm_vecs,
                                                layer_gm_distance[:, np.newaxis]
                                        ) * 0.1
        associated_white_matter_points = wm_intersec_pts - \
                                         np.multiply(
                                                 # multiply unit-normals by individual wm thickness, then scale to 20%
                                                 center_to_wm_vecs,
                                                 layer_wm_distance[:, np.newaxis]
                                         ) * 0.1

        layer_gm_wm_info[layer_id] = {
            "assoc_gm_points": associated_gray_matter_points,
            "assoc_wm_points": associated_white_matter_points,
            "layer_gm_dist": layer_gm_distance,
            "layer_wm_dist": layer_wm_distance
        }

    return layer_gm_wm_info


def calc_e_in_midlayer_roi(
        e,
        roi,
        mesh=None,
        qoi=None,
        layer_gm_wm_info=None
):
    """
    This is to be called by Simnibs as postprocessing function per FEM solve.

    Parameters
    ----------
    e : np.ndarray or tuple
        E to interpolate. Used to be (v, dAdt).
    roi : pynibs.roi.RegionOfInterestSurface
    mesh : simnibs.msh.Mesh
    qoi : list of str
        List of identifiers of the to-be calculated quantities of interest.
    layer_gm_wm_info : dict[str, dict[str, np.ndarray]], optional
        For each layer defined in the mesh (outer dict key: layer_id),
        provide pre-computed geometrical information as a dictionary (outer dict value):

        * key: gm_nodes, For each layer node, the corresponding point on the GM surface.
        * key: wm_nodes, For each layer node, the corresponding point on the WM surface.
        * key: gm_center_distance, The distance between the layer nodes and the corresponding points on the GM.
        * key: wm_center_distance, The distance between the layer nodes and the corresponding points on the WM.

    Returns
    -------
    (roi.n_tris, 4) : np.vstack((e_mag, e_norm, e_tan, e_angle)).transpose() for the midlayer
        (len(roi.layers)x(roi.layers[idx].surface.n_tris,4)) : np.vstack((e_mag, e_norm, e_tan, e_angle)).transpose()
    """
    # set default return quantities
    if qoi is None:
        qoi = ['E', 'mag', 'norm', 'tan', 'angle']
    qois_midlayer = copy.copy(qoi)
    ret_arr_num_components = len(qoi)

    # special care for E
    if 'E' in qoi:
        ret_arr_num_components += 2  # E is a 3D quantity whose components are stored individually
        qois_midlayer.remove('E')  # we don't need this to be computed, it already is

    qois_layers = []  # the QOIs computed for the cortical layers (not midlayer)
    max_num_elmts = roi.n_tris  # number of midlayer elements to define the shape of the output data structure
    interpolation_surfaces = {  # initialize the ROI surfaces the QOIs will be interpolated onto
        "midlayer": simnibs.Msh(
                nodes=simnibs.Nodes(node_coord=roi.node_coord_mid),
                elements=simnibs.Elements(triangles=roi.node_number_list + 1)
        )
    }

    # check if interpolation must be done for the cortical layers and initialize accordingly
    if len(roi.layers) > 0 and layer_gm_wm_info is not None:
        if len(layer_gm_wm_info.keys()) == len(roi.layers):
            qois_layers = ['mag', 'theta', 'gradient']  # = parameters of the neuronal meanfield model

            for layer in roi.layers:
                # Do the cortical layers have more elements than the midlayer?
                max_num_elmts = np.max((max_num_elmts, layer.surface.elm.node_number_list.shape[0]))
                # Expand shape of output data structure by the number of QOIs computed for this layer
                ret_arr_num_components += len(qois_layers)
                # Add this layer to the interpolation surfaces.
                interpolation_surfaces[layer.id] = layer.surface
        else:
            print("[calc_e_in_midlayer] Computation of parameters for neuronal meanfield model was requested,"
                  "but number of layer information in 'layer_gm_wm_info' does not match the number of layers"
                  "found in the mesh. Skipping computation of neuronal model parameters...")

    # We used to use (v, dadt) but nowadays E is enough
    if isinstance(e, tuple):
        e = e[0]
    # simnibs calls this with empty data so find out about the results dimensions
    if (e == 0).all():
        return np.zeros((max_num_elmts, ret_arr_num_components))

    # calc QOIs (all calculations are performed in the element centers)
    data = simnibs.ElementData(e, name='E', mesh=mesh)

    interp_res_per_surface = dict()
    mesh.elmdata.append(simnibs.ElementData(e, name='E', mesh=mesh))

    for surface_id, surface in interpolation_surfaces.items():
        print(f"Computing {surface_id}")
        barycenter_surface = simnibs.Msh(
                nodes=simnibs.Nodes(node_coord=surface.elements_baricenters().value),
                elements=None
        )

        # determine surface normals in elements
        p1_tri = surface.nodes.node_coord[surface.elm.node_number_list[:, 0] - 1, :]
        p2_tri = surface.nodes.node_coord[surface.elm.node_number_list[:, 1] - 1, :]
        p3_tri = surface.nodes.node_coord[surface.elm.node_number_list[:, 2] - 1, :]

        triangles_normals = np.cross(p2_tri - p1_tri, p3_tri - p1_tri).T
        triangles_normals /= np.linalg.norm(triangles_normals, axis=0)
        triangles_normals = triangles_normals.T

        interpolated = data.interpolate_to_surface(barycenter_surface)
        del barycenter_surface
        qois_calculated = collections.OrderedDict()  # maintain order of calculation with OrderedDict

        if surface_id == "midlayer":
            qois_to_calc = qois_midlayer
            # E is the result of the FEM; all other quantities are secondary QOIs computed from E
            if 'E' in qoi:
                qois_calculated['E'] = interpolated.value
                # qois_calculated['E'] = pynibs.data_nodes2elements(data=qois_calculated['E'],
                #                                                   con=(surface.elm.node_number_list - 1))
        elif surface_id != "midlayer" and len(qois_layers) > 0:
            qois_to_calc = qois_layers
        else:
            continue

        for quant in qois_to_calc:
            if quant == 'mag':
                qois_calculated[quant] = interpolated.norm().value
            elif quant == 'norm':
                qois_calculated[quant] = np.sum(qois_calculated["E"] * triangles_normals, axis=1)
                qois_calculated[quant] *= -1
            elif quant == 'tan':
                e_n = np.sum(qois_calculated["E"] * triangles_normals, axis=1)
                e_mag = np.linalg.norm(qois_calculated["E"], axis=1)
                qois_calculated[quant] = np.sqrt(e_mag ** 2 - e_n ** 2)
            elif quant == 'angle':
                qois_calculated[quant] = interpolated.angle().value
            elif quant == 'theta':  # basically 'angle' but we want it in degrees for the meanfield model
                qois_calculated[quant] = e_field_angle_theta(surface, mesh).value
            elif quant == 'gradient':
                qois_calculated[quant] = e_field_gradient_between_wm_gm(
                        surface,
                        mesh,
                        layer_gm_wm_info[surface_id]["assoc_gm_points"],
                        layer_gm_wm_info[surface_id]["assoc_wm_points"],
                        layer_gm_wm_info[surface_id]["layer_gm_dist"],
                        layer_gm_wm_info[surface_id]["layer_wm_dist"],
                ).value
            else:
                raise ValueError('Invalid quantity: {0}'.format(quant))

            # - wrap 1D data in array
            try:
                if qois_calculated[quant].ndim == 1:
                    qois_calculated[quant] = qois_calculated[quant][:, np.newaxis]
            except KeyError:
                pass

        # OrderedDict has maintained the order of the QOI-list.
        interp_res_per_surface[surface_id] = [*qois_calculated.values()]

    mesh.elmdata.pop()
    del qois_calculated, surface, interpolated, mesh, data, e, roi, triangles_normals
    gc.collect()

    # 0-pad midlayer & layer data to have the same number of rows (hstack not possible otherwise),
    # according to the layer with the highest number of elements.
    for surface_id in interp_res_per_surface.keys():
        num_values_to_be_padded = max_num_elmts - interp_res_per_surface[surface_id][0].shape[0]
        for q in range(len(interp_res_per_surface[surface_id])):
            interp_res_per_surface[surface_id][q] = np.pad(interp_res_per_surface[surface_id][q],
                                                   ((0, num_values_to_be_padded), (0, 0)))

    # Stack all result fields column-wise; each column = 1-result array (or in case of E, a component of E)
    # For midlayer + 5 layers, there will be 21 columns and max_num_elements rows
    # 21 = midlayer(3x for E, 1x each for magE, tanE, normE) + 3*layer(1x each for magE, theta, gradient)
    return np.hstack([np.hstack(surf_data) for surf_data in interp_res_per_surface.values()])


def read_coil_geo(fn_coil_geo):
    """
    Reads a coil .geo file.

    Parameters
    ----------
    fn_coil_geo : str
        Filename of .geo file created from SimNIBS containing the dipole information

    This reads data from lines like this:
        SP(-5.906416407434245, -56.83325018618547, 104.15283927746198){0.0};
        or
        VP(-70.46122751969759, -68.28080005454271, 29.538763084748382){-0.10087956492712635, 0.0, -0.9948986447775034};

    Returns
    -------
    dipole_pos : np.ndarray of float
        (n_dip, 3) Dipole positions ``(x, y, z)``.
    dipole_mag : np.ndarray of float
        (n_dip, 1) Dipole magnitude.
    """
    regex = r"(S|V)P\((.*?)\)\{(.*?)\}"
    with open(fn_coil_geo, 'r') as f:
        dipole_pos = []
        dipole_mag = []
        while f:
            te = f.readline()

            if te == "":
                break

            try:
                _, pos, mag_or_vec = re.findall(regex, te)[0]
                pos = [float(x) for x in pos.split(',')]
                mag_or_vec = [float(f) for f in mag_or_vec.split(', ')]
                if len(mag_or_vec) > 1:
                    mag_or_vec = np.linalg.norm(mag_or_vec)
                # else:
                #     mag_or_vec = mag_or_vec[0]
                dipole_pos.append(pos)
                dipole_mag.append(mag_or_vec)
            except IndexError:
                pass

        dipole_pos = np.vstack(dipole_pos)
        dipole_mag = np.atleast_2d(dipole_mag)

    return dipole_pos, dipole_mag


def check_mesh(mesh, verbose=False):
    """
    Check a simmibs.Mesh for degenerated elements:

        * zero surface triangles
        * zerso volume tetrahedra
        * negative volume tetrahedra

    Parameters
    ----------
    mesh : str or simnibs.Mesh

    Other parameters
    ----------------
    verbose : book, default: False
        Print some verbosity messages.

    Returns
    -------
    zero_tris : np.ndarray
        Element indices for zero surface tris (0-indexed)
    zero_tets : np.ndarray
        Element indices for zero volume tets (0-indexed)
    neg_tets : np.ndarray
        Element indicies for negative volume tets (0-indexed)
    """
    if isinstance(mesh, str):
        mesh = simnibs.read_msh(mesh)
    tris = mesh.elm.node_number_list[mesh.elm.triangles - 1][:, :3]
    points_tri = mesh.nodes[tris]
    tri_area = pynibs.calc_tri_surface(points_tri)
    zero_tris = np.argwhere(np.isclose(tri_area, 0, atol=1e-13))
    if verbose:
        print(f"{len(zero_tris)} zero surface triangles found.")

    tets = mesh.elm.node_number_list[mesh.elm.tetrahedra - 1]
    points_tets = mesh.nodes[tets]
    tets_volume = pynibs.calc_tet_volume(points_tets)
    zero_tets = np.argwhere(np.isclose(tets_volume, 0, atol=1e-13))
    if verbose:
        print(f"{len(zero_tets)} zero volume tetrahedra found.")

    tet_idx = mesh.elm.node_number_list[mesh.elm.tetrahedra - 1]
    vol = pynibs.calc_tet_volume(mesh.nodes.node_coord[tet_idx - 1], abs=False)
    neg_idx = np.argwhere(vol > 0)
    if verbose:
        print(f"{len(neg_idx)} negative tets found.")
    neg_idx_in_full_arr = mesh.elm.tetrahedra[neg_idx] - 1

    return zero_tris, zero_tets + len(mesh.elm.triangles), neg_idx_in_full_arr


def fix_mesh(mesh, verbose=False):
    """
    Fixes simnibs.Mesh by removing any zero surface tris and zero volume tets and by fixing negative volume tets.

    Parameters
    ----------
    mesh : str or simnibs.Mesh
        Filename of mesh or mesh object.

    Other parameters
    ----------------
    verbose : bool, default: False
        Print some verbosity messages.

    Returns
    -------
    fixed_mesh : simnibs.Mesh
    """
    if isinstance(mesh, str):
        mesh = simnibs.read_msh(mesh)

    zero_tris, zero_tets, neg_tets = check_mesh(mesh, verbose=verbose)

    if neg_tets.size:
        mesh.elm.node_number_list[neg_tets, [0, 1, 2, 3]] = mesh.elm.node_number_list[neg_tets, [0, 1, 3, 2]]

    if zero_tris.size:
        mesh = mesh.remove_from_mesh(elements=zero_tris + 1)

    if zero_tets.size:
        mesh = mesh.remove_from_mesh(elements=zero_tets + 1 - zero_tris.size)

    # check again
    zero_tris, zero_tets, neg_tets = check_mesh(mesh)

    if zero_tris.size or zero_tets.size or neg_tets.size:
        warnings.warn(f"Couldn't fix mesh: zero_tris: "
                      f"{zero_tris.size} zero tris, {zero_tets.size} zero_tets, {neg_tets.size} neg_tets left over.")

    return mesh


def smooth_mesh(mesh, output_fn, smooth=.8, approach='taubin', skin_only_output=False, smooth_tissue='skin'):
    """
    Smoothes the skin compartment of a simnibs mesh. Uses one of three trimesh.smoothing approaches.
    Because tetrahedra and triangle share the same nodes, this also smoothes the volume domain.

    Parameters
    ----------
    mesh : str or simnibs.Mesh
    output_fn : str
    smooth : float, default: 0.8
        Smoothing aggressiveness. ``[0, ..., 1]``.
    approach: str, default: 'taubin'
        Which smoothing approach to use. One of (``'taubin'``, ``'laplacian'``, ``'humphrey'``.)
    smooth_tissue : str or list of int, default: 'skin'
        Which tissue type to smooth. E.g. ``'gm'`` or ``[2, 1002]``.

    Other parameters
    ----------------
    skin_only_output : bool, default: True
        If true, a skin only mesh is written out instead of the full mesh.

    Returns
    -------
    <file> : The smoothed mesh.

    .. figure:: ../../doc/images/smooth_mesh.png
       :scale: 50 %
       :alt: Original and smoothed surfaces and volumes.

       Left: original, spiky mesh. Right: smoothed mesh.
    """
    if isinstance(mesh, str):
        mesh = simnibs.mesh_io.read_msh(mesh)

    # only triangles
    mesh_cropped = mesh.crop_mesh(elm_type=2)
    if smooth_tissue == 'skin':
        smooth_tissue = [5, 1005]
    elif smooth_tissue == 'gm':
        smooth_tissue = [2, 1002]
    elif smooth_tissue == 'wm':
        smooth_tissue = [1, 1001]
    elif type(smooth_tissue) == str:
        raise ValueError(f"Don't know smooth_tissue='{smooth_tissue}'.")

    # only triangles + only specific tissue type
    mesh_cropped = mesh_cropped.crop_mesh(smooth_tissue)

    # simnibs node indexing is 1-based
    tri_node_nr = mesh_cropped.elm.node_number_list[mesh_cropped.elm.triangles - 1] - 1
    tri_node_nr = tri_node_nr[:, :3]  # triangles only have 3 dimensions

    assert output_fn.endswith('.msh'), f"Wrong file suffix: {output_fn}. Use .msh"
    assert 0 <= smooth <= 1, f"'smooth={smooth}' parameter must be within  [0,1]. "

    # create a Trimesh object based on the skin tris
    mesh_trimesh = trimesh.Trimesh(vertices=mesh_cropped.nodes.node_coord,
                                   faces=tri_node_nr,
                                   process=False)  # process=False keeps original ordering
    if approach == 'taubin':
        smoothed_mesh = trimesh.smoothing.filter_taubin(mesh_trimesh.copy(), lamb=smooth)  # do smoothing
    elif approach == 'laplacian':
        smoothed_mesh = trimesh.smoothing.filter_laplacian(mesh_trimesh.copy(), lamb=smooth)  # do smoothing
    elif approach == 'humphrey':
        smoothed_mesh = trimesh.smoothing.filter_humphrey(mesh_trimesh.copy(), alpha=smooth)  # do smoothing
    else:
        raise NotImplementedError(f"Approach {approach} not implemented. Use 'taubin', laplacian', or 'humphrey'.")

    # find indices of nodes in original mesh
    ind_nodes = np.in1d(mesh.nodes.node_coord[:, 0], mesh_cropped.nodes.node_coord[:, 0]) + \
                np.in1d(mesh.nodes.node_coord[:, 1], mesh_cropped.nodes.node_coord[:, 1]) + \
                np.in1d(mesh.nodes.node_coord[:, 2], mesh_cropped.nodes.node_coord[:, 2])
    ind_nodes = np.where(ind_nodes)[0]

    # This doesn't work if several nodes are at the same location
    if ind_nodes.shape[0] != mesh_cropped.nodes.node_coord.shape[0]:
        # probably some duplicate nodes
        if mesh_cropped.nodes.node_coord.shape[0] != np.unique(mesh_cropped.nodes.node_coord, axis=0).shape[0]:
            # Some other problem
            raise ValueError("Duplicate nodes found in cropped mesh.")

        # Find the duplicate notes in mesh and remove from ind_nodes list
        unique_elms, dup_elms = np.unique(mesh.nodes.node_coord, axis=0, return_counts=True)
        assert unique_elms.shape[0] == dup_elms.shape[0]

        # go over nodes that have been found twice
        for dup_idx in np.where(dup_elms != 1)[0]:
            found_dup_idx = np.argwhere(np.sum(mesh.nodes.node_coord == unique_elms[dup_idx], axis=1) == 3)
            for i in range(1, len(found_dup_idx)):
                try:
                    dup_2_rem = np.argwhere(ind_nodes == found_dup_idx[i])[0]
                    ind_nodes = np.delete(ind_nodes, dup_2_rem)
                except IndexError:
                    pass

    if skin_only_output:
        # replace smoothed nodes in skin only mesh and write to disk
        mesh_cropped.nodes.node_coord = smoothed_mesh.vertices  # overwrite simnibs Mesh object's nodes
        mesh_cropped.write(output_fn)
    else:
        # replace surface node_number list in full mesh and write to disk
        mesh.nodes.node_coord[ind_nodes, :] = smoothed_mesh.vertices
        mesh.write(output_fn)


def get_opt_mat(folder, roi=0):
    """
    Load optimal coil position/orientation matsimnibs from SimNIBS online FEM.

    Parameter:
    ----------
    folder : str
        Folder with optimization results.
    roi : str or int, default: 0
        Region of interest to read data for.

    Returns
    -------
    opt_matsimnibs : np.ndarray
        Optimal coil position/orientation.

    """
    e_fn = os.path.join(folder, 'e.hdf5')
    coil_pos_fn = os.path.join(folder, 'search_positions.hdf5')
    assert os.path.exists(e_fn)
    assert os.path.exists(coil_pos_fn)

    with h5py.File(e_fn, 'r') as f_e:
        keys = list(f_e[f'e/roi_{roi}'].keys())
        keys = [int(k) for k in keys]
        keys.sort()
        max_val = 0
        best_idx = None
        for k in tqdm.tqdm(keys, total=len(keys), desc="Getting max E"):
            mean_e = f_e[f'e/roi_{roi}/{k:0>4}'][:].mean()
            if mean_e > max_val:
                max_val = mean_e
                best_idx = k
        print(f"Max E {max_val.round(2)} found for idx {best_idx}.")

    with h5py.File(coil_pos_fn, 'r') as coil_f:
        if coil_f['matsimnibs'][:].shape[2] != len(keys):
            warnings.warn(f"{coil_pos_fn} and {e_fn} have different sizes: {len(keys)} vs "
                          f"{coil_f['matsimnibs'][:].shape[2]} "
                          f"coil positions.")
        return coil_f['matsimnibs'][:, :, best_idx]


def get_skin_cortex_distance(mesh, coords, radius=5):
    """
    Computes the skin-cortex distance (SCD).

    Parameters
    ----------
    mesh : str or simnibs.Mesh
        Mesh object or .msh filename.
    coords : np.ndarray
        [3,] x,y,z coordinates to compute skin-cortex-distacnce for.
    radius : float, default: 5
        Spherical radius around ``coords`` to include points in for SCD computation.

    Returns
    -------
    SCD : np.ndarray of float
        Skin-cortex distance.
    elm_in_roi : np.ndarray of int
        Elelement numbers from mesh.elm.elm_number that where used to calculate PCD for.
    """
    from simnibs import read_msh
    if isinstance(mesh, str):
        mesh = read_msh(mesh)

    centers = mesh.elements_baricenters()[:]
    dist = np.linalg.norm(centers - coords, axis=1)
    elm = mesh.elm.elm_number[
        (dist < radius) *
        np.isin(mesh.elm.tag1, [2]) *  # only grey matter
        np.isin(mesh.elm.elm_type, [4])  # only tetrahedra
        ]
    skin = centers[np.isin(mesh.elm.tag1, [1005])]  # skin triangles
    return np.linalg.norm(skin - coords, axis=1).min(), elm



import unittest
import tempfile
import pynibs
import sys
import shutil
import os


class TestSubject(unittest.TestCase):
    """
    Test pynibs.subject.* functions.
    """
    subject_id = 'testsub'
    mesh_names = ['charm_refm1_sm', 'headreco', 'headreco_refined_m1']
    create_subject_fn = "create_subject_testsub.py"
    create_subject_fn_org = os.path.join(pynibs.__testdatadir__, create_subject_fn)

    # get temporary folder and move create_subject.py theree
    folder = tempfile.TemporaryDirectory()
    fn_create_subject = os.path.join(folder.name, create_subject_fn)
    fn_subject_object = os.path.join(folder.name, f"{subject_id}.hdf5")
    shutil.copyfile(create_subject_fn_org, fn_create_subject)

    def test_01_create_subject(self):
        cmd = f"{sys.executable} {self.fn_create_subject}"
        assert os.system(cmd) == 0, f"Creating subject failed"

    def test_02_load_subject(self):
        subject = pynibs.load_subject(self.fn_subject_object)
        for mesh_id in self.mesh_names:
            assert mesh_id in subject.mesh.keys(), f"{mesh_id} not found in {self.fn_subject_object}"

    def test_03_load_subject_h5(self):
        no_hdf5_suffix = os.path.join(self.folder.name, 'test.h5')
        shutil.copyfile(self.fn_subject_object, no_hdf5_suffix)
        with self.assertRaises(NotImplementedError):
            pynibs.load_subject(no_hdf5_suffix)
        subject = pynibs.load_subject(no_hdf5_suffix, filetype='hdf5')
        for mesh_id in self.mesh_names:
            assert mesh_id in subject.mesh.keys(), f"{mesh_id} not found in {self.fn_subject_object}"

    def test_04_save_subject(self):
        subject = pynibs.load_subject(self.fn_subject_object)
        fn_new = self.fn_subject_object.replace('.hdf5', '_new.hdf5')
        pynibs.save_subject(subject_id=self.subject_id,
                            subject_folder=self.folder.name,
                            fname=fn_new,
                            mri_dict=subject.mri,
                            mesh_dict=subject.mesh,
                            roi_dict=subject.roi,
                            exp_dict=subject.exp,
                            ps_dict=subject.ps,
                            overwrite=True,
                            check_file_exist=False,
                            verbose=False)
        subject = pynibs.load_subject(fn_new)
        for mesh_id in self.mesh_names:
            assert mesh_id in subject.mesh.keys(), f"{mesh_id} not found in {self.fn_subject_object}"

    # def test_06_add_stuff(self):
    #     subject = pynibs.load_subject(self.fn_subject_object)
    #     mri = subject.mri[0]
    #     subject.add_mri_info(mri)
    #
    #     exp = {'TMS_localite': subject.exp['TMS_localite']}
    #     exp['fn_coil'] = exp['TMS_localite']['fn_coil'][0][0]
    #     subject.add_experiment_info(exp)

    def test_05_print_subject(self):
        subject = pynibs.load_subject(self.fn_subject_object)
        print(subject)



import h5py
import math
import warnings
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
from functools import partial
from numpy import cross as cycross
from scipy.spatial import Delaunay

import pynibs


def calc_tet_volume(points, abs=True):
    """
    Calculate tetrahedra volumes.

    Parameters
    ----------
    points: np.ndarray
        shape: (n_tets,4,3)

        .. code-block:: sh

           [[[Ax, Ay, Az],
             [Bx, By, Bz],
             [Cx, Cy, Cz],
             [Dx, Dy, Dy]],

            [[Ax, Ay, Az],
             [Bx, By, Bz],
             [Cx, Cy, Cz],
             [Dx, Dy, Dy]],
            ...
           ]

    Returns
    -------
    volume: np.ndarray
        shape: ``(n_tets)``

    Other Parameters
    ----------------
    abs : bool, default: true
        Return magnitude
    """
    if points.ndim == 2:
        points = np.atleast_3d(points).reshape(1, 4, 3)
    if points.ndim != 3:
        raise ValueError(f"Wrong dimensions for points: ({points.shape}). Expected: (n_tets,4,3))")

    a = np.zeros((points.shape[0], 4, 4))
    a[:, :, 3] = 1
    a[:, :, :3] = points
    a = a.swapaxes(1, 2)

    if abs:
        return np.abs(1 / 6 * np.linalg.det(a))
    else:
        return 1 / 6 * np.linalg.det(a)


def calc_tri_surface(points):
    """
    Calculate triangle surface areas.

    Parameters
    ----------
    points : np.ndarray
        (n_triangles,3,3)

    Returns
    -------
    triangle_area : np.ndarray
    """
    a = np.linalg.norm(points[:, 0] - points[:, 1], axis=1)
    b = np.linalg.norm(points[:, 1] - points[:, 2], axis=1)
    c = np.linalg.norm(points[:, 0] - points[:, 2], axis=1)
    s = np.sum((a, b, c), axis=0) / 2
    return (s * (s - a) * (s - b) * (s - c)) ** 0.5


def get_sphere(mesh=None, mesh_fn=None, target=None, radius=None, roi_idx=None, roi=None, elmtype='tris', domain=None):
    """
    Return element idx of elements within a certain distance to provided target.
    Element indices are 0-based (tris and tets start at 0, 'pynibs' style)
    Elements might be 'tris' (default) or 'tets'

    If roi object / idx and mesh fn is provided, the roi is expected to have midlayer information and the roi
    geometry is used.

    Parameters
    ----------
    mesh : pynibs.mesh.mesh_struct.TetrahedraLinear, optional
    mesh_fn : str, optional
        Filename to SimNIBS .msh or pyNIBS .hdf5 mesh file.
    target : np.ndarray of float or list of float
        (3,) X, Y, Z coordinates of target.
    radius : float
        Sphere radius im mm.
    roi_idx : str or int, optional
        ROI name.
    elmtype : str, default: 'tris'
        Return triangles or tetrahedra in sphere around target. One of ('tris', 'tets').

    Returns
    -------
    elms_in_sphere : np.ndarray
        (n_elements): Indices of elements found in ROI
    """
    # let's handle the input parameter combinations
    assert target is not None
    assert mesh is not None or mesh_fn is not None
    if elmtype.lower().startswith('tri'):
        elmtype = "tris"
    elif elmtype.lower().startswith('tet'):
        elmtype = "tets"

    if mesh_fn is not None:
        if mesh is not None:
            raise ValueError("Either provide mesh or mesh_fn")
        if mesh_fn.endswith('.hdf5'):
            mesh = pynibs.load_mesh_hdf5(mesh_fn)
        elif mesh_fn.endswith('.msh'):
            mesh = pynibs.load_mesh_msh(mesh_fn)

    if roi is None and roi_idx is not None:
        if mesh_fn is None:
            raise ValueError("Provide mesh_fn to load roi from.")
        roi = pynibs.load_roi_surface_obj_from_hdf5(mesh_fn)[roi_idx]
    if roi is None and roi_idx is not None or roi is not None and roi_idx is None:
        raise ValueError("Provide either roi and roi_idx or none of them.")
    if elmtype == 'tris':
        return tris_in_sphere(mesh=mesh, target=target, radius=radius, roi=roi)
    elif elmtype == 'tets':
        return tets_in_sphere(mesh=mesh, target=target, radius=radius, roi=roi, domain=domain)
    else:
        raise ValueError(f"Unknown elmtype '{elmtype}'")


def tets_in_sphere(mesh, target, radius, roi, domain=None):
    """
    Worker function for get_sphere()

    Returns element idx of elements within a certain distance to provided target.
    If roi object / idx and mesh fn is provided, the roi is expected to have midlayer information and the roi
    geometry is used.

    If radius is None or 0, the nearest element is returned.

    Parameters
    ----------
    mesh : pynibs.TetrahedraLinear, optional
    target : np.ndarray of float, optional
        (3,) X, Y, Z coordinates of target
    radius : float, optional
        Sphere radius im mm
    roi : pynibs.mesh.ROI, optional
        Region of interest

    Returns
    -------
    tets_in_sphere : np.ndarray
        (n_tets): Indices of elements found in ROI

    """
    if roi is None:
        if radius is None or radius == 0:
            return np.where(np.linalg.norm(mesh.tetrahedra_center - target, axis=1) ==
                            np.min(np.linalg.norm(mesh.tetrahedra_center - target, axis=1)))[0]

        else:
            if domain is None:
                domain = [1, 2, 3, 4, 5]
            tet_target_idx = np.where(np.linalg.norm(mesh.tetrahedra_center - target, axis=1) <= radius)[0]
            return np.array([idx for idx in tet_target_idx if mesh.tetrahedra_regions[idx] in domain])

    else:
        warnings.warn("Sphere 'tets' extraction from ROI untested! Consider extracting 'tris' instead.")
        if radius is not None and radius > 0:
            tri_target_idx = np.where(np.linalg.norm(roi.tri_center_coord_mid - target, axis=1) <= radius)[0]
        else:
            tri_target_idx = np.where(np.linalg.norm(roi.tri_center_coord_mid - target, axis=1) == np.min(
                    np.linalg.norm(roi.tri_center_coord_mid - target, axis=1)))[0]
        tet_target_idx = roi.tet_idx_tri_center_mid[tri_target_idx]
        return tet_target_idx


def tris_in_sphere(mesh, target, radius, roi):
    """
    Worker function for get_sphere().

    Returns triangle idx of elements within a certain distance to provided target.
    If roi object / idx and mesh fn is provided, the roi is expected to have midlayer information and the roi
    geometry is used.

    If radius is None or 0, the nearest element is returned.

    Parameters
    ----------
    mesh : pynibs.mesh.TetrahedraLinear, optional
    target : np.ndarray of float or list of float
        (3,) X, Y, Z coordinates of target
    radius : float
        Sphere radius im mm
    roi : pynibs.mesh.mesh_struct.ROI, optional
        ROI

    Returns
    -------
    tris_in_sphere : np.ndarray
        (n_triangles): Indices of elements found in sphere
    """
    if roi is None:
        if radius is None or radius == 0:
            tri_target_idx = np.where(np.linalg.norm(mesh.triangles_center - target, axis=1) ==
                                      np.min(np.linalg.norm(mesh.triangles_center - target, axis=1)))[0]
        else:
            tri_target_idx = np.where(np.linalg.norm(mesh.triangles_center - target, axis=1) <= radius)[0]
    else:
        if radius is not None and radius > 0:
            tri_target_idx = np.where(np.linalg.norm(roi.tri_center_coord_mid - target, axis=1) <= radius)[0]
        else:
            tri_target_idx = np.where(np.linalg.norm(roi.tri_center_coord_mid - target, axis=1) == np.min(
                    np.linalg.norm(roi.tri_center_coord_mid - target, axis=1)))[0]

    return tri_target_idx


def sample_sphere(n_points, r):
    """
    Creates n_points evenly spread in a sphere of radius r.

    Parameters
    ----------
    n_points: int
        Number of points to be spread, must be odd.
    r: float
        Radius of sphere.

    Returns
    -------
    points: np.ndarray of float
        (N x 3), Evenly spread points in a unit sphere.
    """

    assert n_points % 2 == 1, "The number of points must be odd"
    points = []

    # The golden ratio
    phi = (1 + math.sqrt(5)) / 2.
    n = int((n_points - 1) / 2)

    for i in range(-n, n + 1):
        lat = math.asin(2 * i / n_points)
        lon = 2 * math.pi * i / phi
        x = r * math.cos(lat) * math.cos(lon)
        y = r * math.cos(lat) * math.sin(lon)
        z = r * math.sin(lat)
        points.append((x, y, z))

    points = np.array(points, dtype=float)

    return points


def get_indices_discontinuous_data(data, con, neighbor=False, deviation_factor=2,
                                   min_val=None, not_fitted_elms=None, crit='median', neigh_style='point'):
    """
    Get element indices (and the best neighbor index), where the data is discontinuous

    Parameters
    ----------
    data : np.ndarray of float [n_data]
        Data array to analyze given in the element center
    con : np.ndarray of float [n_data, 3 or 4]
        Connectivity matrix
    neighbor : bool, default: False
        Return also the element index of the "best" neighbor (w.r.t. median of data)
    deviation_factor : float
        Allows data deviation from 1/deviation_factor < data[i]/median < deviation_factor
    min_val : float, optional
        If given, only return elements which have a neighbor with data higher than min_val.
    not_fitted_elms : np.ndarray
        If given, these elements are not used as neighbors
    crit: str, default: median
        Criterium for best neighbor. Either median or max value
    neigh_style : str, default: 'point'
        Should neighbors share point or 'edge'

    Returns
    -------
    idx_disc : list of int [n_disc]
        Index list containing the indices of the discontinuous elements
    idx_neighbor : list of int [n_disc]
        Index list containing the indices of the "best" neighbors of the discontinuous elements
    """

    n_ele = con.shape[0]
    idx_disc, idx_neighbor = [], []

    data[data == 0] = 1e-12

    if neigh_style == 'point':
        def get_neigh(m):
            return np.logical_and(0 < mask, mask < 3)
    elif neigh_style == 'edge':
        def get_neigh(m):
            return mask == 2
    else:
        raise NotImplementedError(f"neigh_style {neigh_style} unknown.")

    if crit == 'median':
        def is_neigh():
            if not (1 / deviation_factor < data_i / median < deviation_factor):
                neighbor_indices = np.where(mask_neighbor)[0]
                best_neigh = neighbor_indices[(np.abs(data[neighbor_indices] - median)).argmin()]
                if min_val is None or data[best_neigh] > min_val:
                    idx_disc.append(elm_i)
                # if neighbor:
                idx_neighbor.append(best_neigh)
    elif crit == 'max':
        def is_neigh():
            if data_i / median < 1 / deviation_factor:
                neighbor_indices = np.where(mask_neighbor)[0]
                best_neigh = neighbor_indices[(data[neighbor_indices]).argmax()]
                if min_val is None or data[best_neigh] > min_val:
                    idx_disc.append(elm_i)

                # if neighbor:
                idx_neighbor.append(best_neigh)
    elif crit == 'randmax':
        def is_neigh():
            if data_i / median < 1 / deviation_factor:
                neighbor_indices = np.where(mask_neighbor)[0]
                best_neigh = np.random.choice(neighbor_indices[(data[neighbor_indices]) > 0], 1)
                if min_val is None or data[best_neigh] > min_val:
                    idx_disc.append(elm_i)

                # if neighbor:
                idx_neighbor.append(best_neigh)
    else:
        raise NotImplementedError(f"Criterium {crit} unknown. ")

    for elm_i, data_i in zip(range(n_ele), data):
        if elm_i in not_fitted_elms:
            continue

        # find neighbors
        mask = np.sum(np.isin(con, con[elm_i, :]), axis=1)
        mask_neighbor = get_neigh(mask)

        # best_values are set to 0 for bad elements and unfittable ones. do not use these as neighbors
        if not_fitted_elms is not None and len(not_fitted_elms) != 0:
            mask_neighbor[not_fitted_elms] = False

        # if the element is lonely floating and has no neighbors ... continue
        if not np.sum(mask_neighbor):
            continue

        # check if current value does not fit to neighbors
        median = np.median(data[mask_neighbor])

        if not median:
            median = 1e-12

        is_neigh()

        # if not (1 / deviation_factor < data_i / median < deviation_factor):
        #     # find best neighbor idx
        #     neighbor_indices = np.where(mask_neighbor)[0]
        #
        #     if crit == 'max':
        #         best_neigh = neighbor_indices[(data[neighbor_indices]).argmax()]
        #     elif crit == 'median':
        #         best_neigh = neighbor_indices[(np.abs(data[neighbor_indices] - median)).argmin()]
        #     else:
        #         raise  NotImplementedError(f"Criterium {crit} unknown. ")
        #     if min_val is None or data[best_neigh] > min_val:
        #         idx_disc.append(elm_i)
        #
        #     if neighbor:
        #         idx_neighbor.append(best_neigh)

        # stop = time.time()
        # print(stop-start)

    if neighbor:
        return idx_disc, idx_neighbor
    else:
        return idx_disc


def find_nearest(array, value):
    """
    Given an "array", and given a "value" , returns an index j such that "value" is between array[j]
    and array[j+1]. "array" must be monotonic increasing. j=-1 or j=len(array) is returned
    to indicate that "value" is out of range below and above respectively.

    Parameters
    ----------
    array : np.ndarray of float
        Monotonic increasing array.
    value : float
        Target value the nearest neighbor index in ``array`` is computed for.

    Returns
    -------
    idx : int
        Index j such that "value" is between array[j] and array[j+1].

    """
    n = len(array)
    if value < array[0]:
        return -1
    elif value > array[n - 1]:
        return n
    jl = 0  # Initialize lower
    ju = n - 1  # and upper limits.
    while ju - jl > 1:  # If we are not yet done,
        jm = (ju + jl) >> 1  # compute a midpoint with a bitshift
        if value >= array[jm]:
            jl = jm  # and replace either the lower limit
        else:
            ju = jm  # or the upper limit, as appropriate.
    # Repeat until the test condition is satisfied.
    if value == array[0]:  # edge cases at bottom
        return 0
    elif value == array[n - 1]:  # and top
        return n - 1
    else:
        return jl


def in_hull(points, hull):
    """
    Test if points in `points` are in `hull`.
    `points` should be a [N x K] coordinates of N points in K dimensions.
    `hull` is either a scipy.spatial.Delaunay object or the [M x K] array of the
    coordinates of M points in Kdimensions for which Delaunay triangulation
    will be computed.

    Parameters
    ----------
    points : np.ndarray
        (N_points x 3) Set of floating point data to test whether they are lying inside the hull or not.
    hull : scipy.spatial.Delaunay or np.ndarray
         (M x K) Surface data.

    Returns
    -------
    inside : np.ndarray of bool
        TRUE: point inside the hull
        FALSE: point outside the hull
    """

    if not isinstance(hull, Delaunay):
        hull = Delaunay(hull)
    return hull.find_simplex(points) >= 0


def calc_tetrahedra_volume_cross(P1, P2, P3, P4):
    """
    Calculates volume of tetrahedra specified by the 4 points P1...P4
    multiple tetrahedra can be defined by P1...P4 as 2-D np.ndarrays
    using the cross and vector dot product.

    .. math::
        P1=\\begin{bmatrix}
        x_{{tet}_1} & y_{{tet}_1} & z_{{tet}_1}   \\\\
        x_{{tet}_2} & y_{{tet}_2} & z_{{tet}_2}   \\\\
        ... & ... & ...    \\\\
        x_{{tet}_N} & y_{{tet}_N} & z_{{tet}_N}    \\\\
        \\end{bmatrix}

    Parameters
    ----------
    P1 : np.ndarray of float [N_tet x 3]
        Coordinates of first point of tetrahedra
    P2 : np.ndarray of float [N_tet x 3]
        Coordinates of second point of tetrahedra
    P3 : np.ndarray of float [N_tet x 3]
        Coordinates of third point of tetrahedra
    P4 : np.ndarray of float [N_tet x 3]
        Coordinates of fourth point of tetrahedra

    Returns
    -------
    tetrahedra_volume: np.ndarray of float [N_tet x 1]
        Volumes of tetrahedra
    """

    tetrahedra_volume = 1.0 / 6 * \
                        np.sum(np.multiply(cycross(P2 - P1, P3 - P1), P4 - P1), 1)
    tetrahedra_volume = tetrahedra_volume[:, np.newaxis]
    return tetrahedra_volume


def calc_tetrahedra_volume_det(P1, P2, P3, P4):
    """
    Calculate volume of tetrahedron specified by 4 points P1...P4
    multiple tetrahedra can be defined by P1...P4 as 2-D np.arrays
    using the determinant.


    .. math::
        P1=\\begin{bmatrix}
        x_{{tet}_1} & y_{{tet}_1} & z_{{tet}_1}   \\\\
        x_{{tet}_2} & y_{{tet}_2} & z_{{tet}_2}   \\\\
        ... & ... & ...    \\\\
        x_{{tet}_N} & y_{{tet}_N} & z_{{tet}_N}    \\\\
        \\end{bmatrix}

    Parameters
    ----------
    P1 : np.ndarray of float [N_tet x 3]
        Coordinates of first point of tetrahedra
    P2 : np.ndarray of float [N_tet x 3]
        Coordinates of second point of tetrahedra
    P3 : np.ndarray of float [N_tet x 3]
        Coordinates of third point of tetrahedra
    P4 : np.ndarray of float [N_tet x 3]
        Coordinates of fourth point of tetrahedra

    Returns
    -------
    tetrahedra_volume : np.ndarray of float [N_tet x 1]
        Volumes of tetrahedra
    """

    N_tets = P1.shape[0] if P1.ndim > 1 else 1

    # add ones
    j1 = np.hstack((np.ones((N_tets, 1)), P1))
    j2 = np.hstack((np.ones((N_tets, 1)), P2))
    j3 = np.hstack((np.ones((N_tets, 1)), P3))
    j4 = np.hstack((np.ones((N_tets, 1)), P4))

    j = np.zeros((P1.shape[0] if P1.ndim > 1 else 1, 4, 4))

    j[:, :, 0] = j1
    j[:, :, 1] = j2
    j[:, :, 2] = j3
    j[:, :, 3] = j4

    tetrahedra_volume = 1.0 / 6.0 * np.linalg.det(j)
    tetrahedra_volume = tetrahedra_volume[:, np.newaxis]
    return tetrahedra_volume


def calc_gradient_surface(phi, points, triangles):
    """
    Calculate gradient of potential phi on surface (i.e. tangential component) given in vertices of a triangular
    mesh forming a 2D surface.

    Parameters
    ----------
    phi : np.ndarray of float [N_points x 1]
        Potential in nodes
    points : np.ndarray of float [N_points x 3]
        Coordinates of nodes (x,y,z)
    triangles : np.ndarray of int32 [N_tri x 3]
        Connectivity of triangular mesh

    Returns
    -------
    grad_phi : np.ndarray of float [N_tri x 3]
        Gradient of potential phi on surface
    """

    grad_phi = np.zeros((triangles.shape[0], 3))

    for i in range(triangles.shape[0]):
        a = np.array([[points[triangles[i, 0], 0] - points[triangles[i, 2], 0],
                       points[triangles[i, 0], 1] - points[triangles[i, 2], 1],
                       points[triangles[i, 0], 2] - points[triangles[i, 2], 2]],
                      [points[triangles[i, 1], 0] - points[triangles[i, 2], 0],
                       points[triangles[i, 1], 1] - points[triangles[i, 2], 1],
                       points[triangles[i, 1], 2] - points[triangles[i, 2], 2]]])

        b = np.array([phi[triangles[i, 0]] - phi[triangles[i, 2]],
                      phi[triangles[i, 1]] - phi[triangles[i, 2]]])

        grad_phi[i, :] = np.dot(np.linalg.pinv(a), b).T

    return grad_phi


def determine_e_midlayer_workhorse(fn_e_results, subject, mesh_idx, midlayer_fun, fn_mesh_hdf5, roi_idx, phi_scaling=1.,
                                   verbose=False):
    """
    phi_scaling: float
        simnibs < 3.0  : 1000.
        simnibs >= 3.0 :    1. (Default)
    """

    if verbose:
        print(f"Loading Mesh and ROI {roi_idx} from {fn_mesh_hdf5}")

    msh = pynibs.load_mesh_hdf5(fn_mesh_hdf5)
    roi = pynibs.load_roi_surface_obj_from_hdf5(fn_mesh_hdf5)

    for fn_e in fn_e_results:

        with h5py.File(fn_e + ".hdf5", 'r') as f:
            phi = f['data/nodes/v'][:][:, np.newaxis]
            # phi = f['data/potential'][:][:, np.newaxis]
            dadt = f['data/nodes/D'][:]
            # dadt = np.reshape(f['data/dAdt'][:], (phi.shape[0], 3), order="c")

        # determine e_norm and e_tan for every simulation
        if verbose:
            print(f"Determine midlayer E-field for {fn_e}.hdf5")

        # choose which function to use for midlayer computation
        if midlayer_fun == "pynibs":
            e_norm_temp, e_tan_temp = msh.calc_E_on_GM_WM_surface3(phi=phi * phi_scaling,
                                                                   dAdt=dadt,
                                                                   roi=roi[roi_idx],
                                                                   verbose=False,
                                                                   mode='magnitude')

            e_norm_temp = e_norm_temp.flatten() * -1
            e_tan_temp = e_tan_temp.flatten()
            e_mag_temp = np.linalg.norm(np.vstack([e_norm_temp, e_tan_temp]).transpose(), axis=1).flatten()

        elif midlayer_fun == "simnibs":
            e_norm_temp_simnibs, e_tan_temp_simnibs = msh.calc_E_on_GM_WM_surface_simnibs_KW(phi=phi * phi_scaling,
                                                                                             dAdt=dadt,
                                                                                             roi=roi[roi_idx],
                                                                                             verbose=False,
                                                                                             subject=subject,
                                                                                             mesh_idx=mesh_idx)

            e_norm_temp_simnibs = e_norm_temp_simnibs.flatten()
            e_tan_temp_simnibs = e_tan_temp_simnibs.flatten()
            e_mag_temp_simnibs = np.linalg.norm(np.vstack([e_norm_temp_simnibs, e_tan_temp_simnibs]).transpose(),
                                                axis=1).flatten()
        else:
            raise ValueError(f"midlayer_fun {midlayer_fun} not implemented.")

        del phi, dadt

        with h5py.File(fn_e + ".hdf5", 'a') as f:
            try:
                del f['data/midlayer/roi_surface/{}/E_mag'.format(roi_idx)]
                del f['data/midlayer/roi_surface/{}/E_tan'.format(roi_idx)]
                del f['data/midlayer/roi_surface/{}/E_norm'.format(roi_idx)]
            except KeyError:
                pass

            f.create_dataset('data/midlayer/roi_surface/{}/E_mag'.format(roi_idx), data=e_mag_temp_simnibs)
            f.create_dataset('data/midlayer/roi_surface/{}/E_tan'.format(roi_idx), data=e_tan_temp_simnibs)
            f.create_dataset('data/midlayer/roi_surface/{}/E_norm'.format(roi_idx), data=e_norm_temp_simnibs)

        if verbose:
            print("\tAdding results to {}".format(fn_e + ".hdf5"))


def determine_e_midlayer(fn_e_results, fn_mesh_hdf5, subject, mesh_idx, roi_idx, n_cpu=4, midlayer_fun="simnibs",
                         phi_scaling=1., verbose=False):
    """
    Parallel version to determine the midlayer e-fields from a list of .hdf5 results files

    Parameters
    ----------
    fn_e_results : list of str
        List of results filenames (.hdf5 format)
    fn_mesh_hdf5 : str
        Filename of corresponding mesh file
    subject : pynibs.Subject
        Subject object
    mesh_idx : int
        Mesh index
    roi_idx : int
        ROI index
    n_cpu : int, default: 4
        Number of parallel computations
    midlayer_fun : str, default: "simnibs"
        Method to determine the midlayer e-fields ("pynibs" or "simnibs")
    phi_scaling : float, default: 1.0
        Scaling factor of scalar potential to change between "m" and "mm"

    Returns
    -------
    <File> .hdf5 file
        Adds midlayer e-field results to ROI
    """

    # msh = pynibs.load_mesh_msh(subject.mesh[mesh_idx]['fn_mesh_msh'])

    n_cpu_available = multiprocessing.cpu_count()
    n_cpu = min(n_cpu, n_cpu_available)

    workhorse_partial = partial(determine_e_midlayer_workhorse,
                                subject=subject,
                                mesh_idx=mesh_idx,
                                midlayer_fun=midlayer_fun,
                                fn_mesh_hdf5=fn_mesh_hdf5,
                                roi_idx=roi_idx,
                                phi_scaling=phi_scaling,
                                verbose=verbose)

    fn_e_results_chunks = pynibs.compute_chunks(fn_e_results, n_cpu)
    pool = multiprocessing.Pool(n_cpu)
    pool.map(workhorse_partial, fn_e_results_chunks)
    pool.close()
    pool.join()


def find_element_idx_by_points(nodes, con, points):
    """
    Finds the tetrahedral element index of an arbitrary point in the FEM mesh.

    Parameters
    ----------
    nodes : np.ndarray [N_nodes x 3]
        Coordinates (x, y, z) of the nodes
    con : np.ndarray [N_tet x 4]
        Connectivity matrix
    points : np.ndarray [N_points x 3]
        Points for which the element indices are found.

    Returns
    -------
    ele_idx : np.ndarray [N_points]
        Element indices of tetrahedra where corresponding 'points' are lying in
    """

    node_idx = []
    for i in range(points.shape[0]):
        node_idx.append(np.where(np.linalg.norm(nodes - points[i, :], axis=1) < 1e-2)[0])

    # ele_idx = np.where((con == np.array(node_idx)).all(axis=1))[0]
    ele_idx = np.where(np.all(np.sort(con, axis=1) == np.sort(np.array(node_idx).flatten()), axis=1))[0]
    return ele_idx


def check_islands_for_single_elm(source_elm, connectivity=None, adjacency=None, island_crit=1):
    """
    This identifies islands in a mesh for a given element. An island is a set of elements, that is only connect
    via a single node to another set of elements.
    These islands usually crash the FEM solver and should be removed.

    1. Find all elements connect to source_elm via one node (1-node-neighbor)
    2. Start with source_elm and visit all 2-node-neighbors ('shared-edge)
    3. Continue recursively with all 2-node-neighbors and visit their 2-node-neighbors
    4. See if any 1-node-neighbors have not been visited with this strategy. If so, an island has been found

    Parameters
    ----------
    source_elm : int
        The source element to check
    connectivity : np.ndarray, optional
        Connectivity ('node_number_list') starting with 0. Can be triangles or tetrahedra
        (n_elms, 3) or (n_elms_4).
    adjacency : np.ndparray, optional
        Adjenceny matrix (n_elm, n_elm). Weights are supposed to be number of shared nodes.
        Computed from neighbors if not provided.
    island_crit : int, default: 'any'
        How many nodes to define islands?
        'any' -> Elements connected via a single node or single edge are defined as an island.
        'node' -> Elements connected via a single _node_ are defined as an island.
        'edge' -> Elements connected via a single _edge_ are defined as an island.

    Returns
    -------
    n_visited : int
    n_not_visited : int
    neighbors_visited : dict, which neighbors have been visited and which have not
    """
    if adjacency is not None and connectivity is not None:
        raise ValueError(f"Provide either neighbors or connectivity, not both.")

    if adjacency is None:
        assert connectivity is not None
        adjacency = np.array([np.sum(np.isin(connectivity, elm), axis=1) for elm in connectivity])

    if island_crit == 'any':
        # find elements with only one node in common
        neighbors_visited = {i: False for i in np.where(adjacency[source_elm] >= 1)[0]}
        island_crit = 1
    elif island_crit == 'node':
        # find elements with only one node in common
        neighbors_visited = {i: False for i in np.where(adjacency[source_elm] == 1)[0]}
        island_crit = 1
    elif island_crit == 'edge':
        # find elements with only one node in common
        neighbors_visited = {i: False for i in np.where(adjacency[source_elm] == 2)[0]}
        island_crit = 2
    else:
        raise ValueError

    # now visit all elements with 2 or more neighboring elements recursivly.
    # everything that's left over between 1 and 2 neighbors is an island
    neighs_to_check = set(np.where(adjacency[source_elm] >= island_crit)[0].tolist())

    # add the starting element to the list
    neighs_to_check.add(source_elm)
    # go through all elements in list
    # print(neighs_to_check)
    while neighs_to_check:
        elm = neighs_to_check.pop()
        neighbors_visited[elm] = True

        # now add all 2-node neighbors for this element
        # for neigh in np.where(adjacency[elm] > 1)[0]:
        sort = (-adjacency[elm]).argsort()
        for idx in np.arange((np.bincount(adjacency[elm])[(island_crit):]).sum()):
            i = sort[idx]
            if i not in neighbors_visited or not neighbors_visited[i]:
                # print(i)
                neighs_to_check.add(i)

    return np.sum([v for v in neighbors_visited.values()]), \
        np.sum([not v for v in neighbors_visited.values()]), \
        neighbors_visited


def find_islands(connectivity=None, adjacency=None, island_crit='any', verbose=False, largest=False):
    """
    This identifies islands in a mesh. An island is a set of elements, that is only connect
    via a single node to another set of elements.
    These islands usually crash the FEM solver and should be removed.

    For each element:
        1. Find all elements connect to source_elm via one node (1-node-neighbor)
        2. Start with source_elm and visit all 2-node-neighbors ('shared-edge)
        3. Continue recursively with all 2-node-neighbors and visit their 2-node-neighbors
        4. See if any 1-node-neighbors have not been visited with this strategy. If so, an island has been found

    .. figure:: ../../doc/images/find_islands.png
       :scale: 50 %
       :alt: Island detection

       Islands are groups of elements that are only connected via a single node/edge to another group.

    Parameters
    ----------
    connectivity : np.ndarray, optional
        Connectivity ('node_number_list') starting with 0. Can be triangles or tetrahedra
        (n_elms, 3) or (n_elms_4).
    adjacency : np.ndparray, optional
        Adjenceny matrix (n_elm, n_elm). Weights are supposed to be number of shared nodes.
        Computed from neighbors if not provided.
    island_crit : int or str, default: 'any'
        How many nodes to define islands?
        'any' -> Elements connected via a single node or single edge are defined as an island.
        'node' -> Elements connected via a single _node_ are defined as an island.
        'edge' -> Elements connected via a single _edge_ are defined as an island.
    largest : book, default: False
        Only return largest island, speeds up computation quite a bit if only one large, and many small islands exist.
    verbose : bool, optional
        Print some verbosity information. Default: False
    Returns
    -------
        elms_with_island : list
            Elements with neighboring islands
        counter_visited : np.ndarray
            shape = (n_elms). How often as each element been visited.
        counter_not_visited : np.ndarray
            shape = (n_elms). How often as each element not been visited.
    """
    elms_with_island = []
    if adjacency is not None and connectivity is not None:
        raise ValueError(f"Provide either neighbors or connectivity, not both.")
    if adjacency is None and connectivity is None:
        raise ValueError(f"Provide either neighbors or connectivity")
    if adjacency is None:
        assert connectivity is not None
        adjacency = np.array([np.sum(np.isin(connectivity, elm), axis=1) for elm in connectivity])

    counter_not_visited = np.zeros(adjacency.shape[0])
    counter_visited = np.zeros(adjacency.shape[0])

    size_island = 0
    visited_elms = set()
    # go through elements and check for islands
    for elm_source in tqdm(range(adjacency.shape[0]), desc="Checking for islands."):
        if largest and elm_source in visited_elms:
            continue
        _, n_not_visited, visited = check_islands_for_single_elm(elm_source,
                                                                 adjacency=adjacency,
                                                                 island_crit=island_crit)

        # add stats for each element how often it has been visited
        elms_visited = [i for i, v in visited.items() if v]

        if not largest:
            counter_visited[elms_visited] += 1
            elms_not_visited = [i for i, v in visited.items() if not v]
            counter_not_visited[elms_not_visited] += 1
        else:
            visited_elms.update(elms_visited)
            if len(elms_visited) > size_island:
                elms_not_visited = [i for i, v in visited.items() if not v]
                size_island = len(elms_visited)
                visited_elms.update(elms_visited)
                counter_visited = np.zeros(adjacency.shape[0])
                counter_not_visited = np.zeros(adjacency.shape[0])
                counter_visited[elms_visited] += 1
                counter_not_visited[elms_not_visited] += 1

        # if an island is found add it to results list
        if n_not_visited:
            if verbose:
                print(f"\nElement {elm_source: >4}: {n_not_visited} 1-node-neighbors not visited. ")
            elms_with_island.append(elm_source)

    return elms_with_island, counter_visited, counter_not_visited


def find_island_elms(connectivity=None, adjacency=None, verbose=False, island_crit='edge', decision='cumulative'):
    """
    Searches for islands in a mesh and returns element indices of the smallest island.
    Island is defines as a set of elements, which share a single node and/or single edge with the rest of the mesh.

    Parameters
    ----------
    connectivity : np.ndarray, optional
        Connectivity ('node_number_list') starting with 0. Can be triangles or tetrahedra
        (n_elms, 3) or (n_elms_4).
    adjacency : np.ndparray, optional
        Adjenceny matrix (n_elm, n_elm). Weights are supposed to be number of shared nodes.
        Computed from neighbors if not provided.
    island_crit : int, default: 'edge'
        How many nodes to define islands?
        'node' -> Elements connected via a single _node_ are defined as an island.
        'edge' -> Elements connected via a single _edge_ are defined as an island.
    decision : str, default: cumulative
        'cumulative' -> Return all element indices that are not visited any times
        'smallest'   -> Return smallest island.
    verbose : bool, optional
        Print some verbosity information. Default: False

    Returns
    -------
        island : list of island-elms
    """
    if adjacency is not None and connectivity is not None:
        raise ValueError(f"Provide either neighbors or connectivity, not both.")
    if adjacency is None:
        assert connectivity is not None
        adjacency = np.array([np.sum(np.isin(connectivity, elm), axis=1) for elm in connectivity])

    all_islands, counter_visited, counter_not_visited = find_islands(connectivity, verbose=verbose,
                                                                     island_crit=island_crit)

    if decision == 'smallest':
        # find the size of the islands
        one_node_neighs = {k: np.where(adjacency[k] == island_crit)[0] for k in all_islands}

        visited, not_visited = {}, {}
        for island, one_node_neigh in one_node_neighs.items():

            # two counters
            n_visited_i, n_not_visited_i = 0, 0

            for island_i in one_node_neigh:
                n_visited, n_not_visited, _ = check_islands_for_single_elm(island_i, adjacency=adjacency,
                                                                           island_crit=island_crit + 1)
                n_visited_i += n_visited
                n_not_visited_i += n_not_visited
            visited[island] = n_visited_i
            not_visited[island] = n_not_visited_i

        # choose the smallest island and get all 2-neighbors
        smallest_island_idx = np.argmin(list(visited.values()))
        smallest_island = list(visited.keys())[smallest_island_idx]

        _, _, elm_idx_from_smalles_island = check_islands_for_single_elm(smallest_island, island_crit=island_crit + 1,
                                                                         adjacency=adjacency)

        if verbose:
            print(f"Island with {len(list(elm_idx_from_smalles_island.keys()))} elements found.")
        return list(elm_idx_from_smalles_island.keys())

    elif decision == 'cumulative':
        return np.argwhere(counter_not_visited > 0)


def cortical_depth(mesh_fn, geo_fn=None, write_xdmf=True, skin_surface_id=1005, verbose=False):
    """
    Compute skin-cortex-distance (SCD) for surface and volume data in ``mesh_fn``.

    .. figure:: ../../doc/images/cortical_depth.png
       :scale: 50 %
       :alt: Visualized cortical depth.

    Cortical depth computed against skin surface.

    Parameters
    ----------
    mesh_fn : str
        :py:class:`~pynibs.mesh.mesh_struct.TetrahedraLinear` mesh file.
    geo_fn : str, optional
        :py:class:`~pynibs.mesh.mesh_struct.TetrahedraLinear` mesh file with geometric data. If provided, geometric
        information is read from here.
    write_xdmf : bool, default: True
        Write .xdmf or not.
    skin_surface_id : int, default: 1005
        Which tissue type nr to compute distance against.
    verbose : bool, default: False
        Print some verbosity information.

    Returns
    -------
    <file> : .hdf5
        ``mesh_fn`` or ``geo_fn`` with SCD information in ``/data/tris/Cortex_dist`` and ``/data/tets/Cortex_dist``.
    <file> : .xdmf
        Only if ``write_xdmf == True``.
    """
    if geo_fn is None:
        geo_fn = mesh_fn
    with h5py.File(mesh_fn, 'r') as f:
        skin_tri_idx = f['/mesh/elm/tri_tissue_type'][:] == skin_surface_id
        tri_nodes = f['/mesh/elm/triangle_number_list'][:][skin_tri_idx]

    elms_with_island, counter_visited, counter_not_visited = pynibs.find_islands(connectivity=tri_nodes,
                                                                                 verbose=True,
                                                                                 island_crit='edge',
                                                                                 largest=True)
    hdf5 = pynibs.load_mesh_hdf5(mesh_fn)

    with h5py.File(geo_fn, 'r') as geo:
        # get indices for skin elements
        skin_tri_idx = np.squeeze(np.argwhere((geo['mesh/elm/tri_tissue_type'][:] == skin_surface_id)))
        skin_positions = hdf5.triangles_center[skin_tri_idx[counter_visited > 0]]

    def fun(row):
        return np.min(np.linalg.norm(row - skin_positions, axis=1))

    if verbose:
        print("Computing triangles")
    distances_tri = np.apply_along_axis(fun, axis=1, arr=hdf5.triangles_center)
    if verbose:
        print("Computing tetrahedra")

    distances_tets = np.apply_along_axis(fun, axis=1, arr=hdf5.tetrahedra_center)
    with h5py.File(mesh_fn, 'a') as f:
        try:
            del f['data/tris/Cortex_dist']
            del f['data/tets/Cortex_dist']
        except KeyError:
            pass
        f.create_dataset(name='data/tris/Cortex_dist', data=distances_tri)
        f.create_dataset(name='data/tets/Cortex_dist', data=distances_tets)

    if write_xdmf:
        pynibs.write_xdmf(overwrite_xdmf=True, hdf5_geo_fn=geo_fn, hdf5_fn=mesh_fn)


def calc_distances(coords, mesh_fn, tissues=None):
    """
    Calculates the distances between ``coords`` and tissue types.

    Parameters
    ----------
    coords : list of list or list or np.ndarray
        Coordinates (X, Y, Z) to compute depths for.
    mesh_fn : str
        pynibs.Mesh hdf5 filename.
    tissues : list of int, optional
        Which tissue types to compute depths for. If none, distances to all tissue types are computed.

    Returns
    -------
    distances : pd.Dataframe()
        colunms: coorrd, tissue_type, distance

    """
    coords = np.atleast_2d(coords)
    print("Coordinate  |  tissue type  | Distance")
    print("=" * 40)
    res = {'coord': [],
           'tissue_type': [],
           'distance': []}
    with h5py.File(mesh_fn, 'r') as f:
        if tissues is None:
            tissues = np.unique(f['mesh/elm/tag2'][:])
        for coord in coords:
            assert np.min(f['mesh/elm/node_number_list'][:]) == 0

            for tissue in tissues:
                if tissue > 100:
                    # if elmtype == 'tris':
                    elmtype = 2
                    node_list = f['mesh/elm/node_number_list'][:][
                        (f['mesh/elm/tag2'][:] == tissue) & (f['mesh/elm/elm_type'][:] == elmtype)]
                    node_list = node_list[:, :3]
                else:
                    elmtype = 4
                    node_list = f['mesh/elm/node_number_list'][:][
                        (f['mesh/elm/tag2'][:] == tissue) & (f['mesh/elm/elm_type'][:] == elmtype)]
                node_coords = f['mesh/nodes/node_coord'][:][node_list]
                distances = np.mean(np.linalg.norm(coord - node_coords, axis=2), axis=1).min()
                res['coord'].append(coord)
                res['tissue_type'].append(tissue)
                res['distance'].append(distances)
                print(f"{coord} |     {tissue: >4} |  {distances.round(2): >6} mm")
            print("-" * 40)
    return pd.DataFrame().from_dict(res)



import pynibs
import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import ticker


try:
    matplotlib.use("Qt5Agg")
except ImportError:
    pass


def plot_io_curve(
        mep_data,
        local_mag_e,
        title="I/O curve",
        y_axis_label="EMG p2p-amplitude, µV",
        x_axis_label="mag(E), V/m, scaled",
        fit_fun=None,
        screenshot_fn=None,
        interactive=False,
        mso=None):
    """
    Plot scattered data of MEP-mag(E) pairs (=> I/O curve).
    Optionally the fit of the proided fit function can be overlaid.

    Parameters
    ----------
    mep_data: np.ndarray[float], [n_stims]
        The acquired MEPs fo each stimulation.
    local_mag_e: np.ndarray[float], [n_stims]
        The local electric field magnitude at the ROI element the I/O curve should be plotted from.
    title: str, optional
        The title of the plot; default: "I/O curve"
    y_axis_label: str, optional
        The title of the y-axis; default: "EMG p2p-amplitude, µV"
    x_axis_label: str, optional
        The title of the x-axis; default: "mag(E), V/m, scaled"
    fit_fun: Callable, optional
        If provided, the fit of the I/O curve with the provided function will be computed and displayed.
        Currently supported: pynibs.simgoid4, pynibs.sigmoid4_log, pynibs.linear
    screenshot_fn: str, optional
        If provided, output the plotted image to that file location.
    interactive: bool
        If True, spawn plot in blocking window.
    mso : np.ndarray or list, optional
        MSO values.

    Returns
    -------
        True if successful, False otherwise.
    """
    if interactive:
        matplotlib.use('Qt5Agg')
    else:
        matplotlib.use('Agg')

    plt.figure(figsize=(11, 11))
    plt.scatter(y=np.log10(mep_data) if fit_fun == pynibs.expio.fit_funs.sigmoid4_log else mep_data,
                x=local_mag_e,
                cmap=plt.cm.get_cmap('spring_r'),
                c=mso if mso is not None else list(range(len(local_mag_e))),
                vmin=0, vmax=100 if mso is not None else len(local_mag_e))

    plt.scatter(y=np.log10(mep_data[-1]) if fit_fun == pynibs.expio.fit_funs.sigmoid4_log else mep_data[-1],
                x=local_mag_e[-1],
                cmap=plt.cm.get_cmap('spring_r'),
                c=mso[-1] if mso is not None else len(local_mag_e),
                s=500,
                vmin=0, vmax=100 if mso is not None else len(local_mag_e))

    if fit_fun is not None:
        local_mag_e_np = np.zeros((len(local_mag_e), 1))
        local_mag_e_np[:, 0] = local_mag_e

        fit_scores, fit_params = pynibs.regression.regress_data(
                e_matrix=np.array(local_mag_e_np),
                mep=mep_data,
                fun=fit_fun,
                n_cpu=16,
                con=None,
                n_refit=10,
                return_fits=True,
                verbose=False,
                pool=None,
                refit_discontinuities=False,
                select_signed_data=False
        )

        num_samples = 100
        x = np.linspace(0, np.min((120, np.max(local_mag_e))), num_samples)
        if fit_fun == pynibs.expio.fit_funs.sigmoid4_log or fit_fun == pynibs.expio.fit_funs.sigmoid4:
            y = fit_fun(x,
                        x0=fit_params[0]['x0'],
                        r=fit_params[0]['r'],
                        amp=fit_params[0]['amp'],
                        y0=fit_params[0]['y0']
                        )
        elif fit_fun == pynibs.expio.fit_funs.linear:
            y = fit_fun(x,
                        m=fit_params[0]["m"],
                        n=fit_params[0]["n"]
                        )
        else:
            y = np.zeros(num_samples)

        title += f", R2: {round(fit_scores[0], 5): <5}"
        plt.plot(x, y, color='black')
        props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
        # for k, v in fit_params.items()
        text = '\n'.join(f'{k: >7}: {np.round(v,2):2.2f}' for k, v in fit_params[0].items())
        text = f'Fitted params:\n‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾\n{text}\n   n_zaps: {len(mep_data): >6}'
        text += f'\n% MSO: {int(mso[-1]): >5}' if mso is not None else ''
        plt.text(local_mag_e.max() - 5, 2, text, fontsize=18,
                 verticalalignment='top', bbox=props, horizontalalignment='right')
        cb = plt.colorbar()
        cb.ax.set_ylabel("% MSO" if mso is not None else "Stimulation Id", fontsize=18)

    if screenshot_fn is not None:
        if fit_fun == pynibs.expio.fit_funs.sigmoid4_log:
            tick_log_rev = lambda x, y: int(pow(10, x))
            ytick_formatter = ticker.FuncFormatter(tick_log_rev)
            y_axis_label += ",\n log10-scaled"
            ax = plt.gca()
            ax.yaxis.set_major_formatter(ytick_formatter)

        plt.title(title, fontsize=32)
        plt.xlabel(x_axis_label, fontsize=30)
        plt.ylabel(y_axis_label, fontsize=30)
        plt.xticks(fontsize=28)
        plt.yticks(fontsize=28)
        plt.savefig(screenshot_fn)

    if interactive:
        plt.show()
    plt.close()



""" Functions to import data from Rogue Research Brainsight neuronavigation systems """
from collections import OrderedDict
import pandas as pd
import numpy as np
import warnings
import datetime
import pynibs
import h5py
import io
import os


class BrainsightCSVParser:
    """
    Parser for Brainsight CSV files.

    Attributes
    ----------
    current_state : BrainsightCSVParser.State
        The current state of the parser.
    header_file : str
        The file header.
    header_data : list of str
        The header data.
    data : list of list of str
        The parsed data rows.

    Methods
    -------
    check_state_transition(line)
        Checks the state transition based on the given line.
    parse_line(line)
        Parses a line and updates the corresponding attributes.
    """

    class State:
        FILE_HEADER = 0
        SKIPPING = 1
        DATA_HEADER = 2
        DATA = 3
        FIN = 4

    def __init__(self):
        """
        Initialize the BrainsightCSVParser object.
        """
        self.current_state = BrainsightCSVParser.State.FILE_HEADER
        self.header_file = ""
        self.header_data = []
        self.data = []

    def check_state_transition(self, line):
        """
        Checks the state transition based on the given line.

        Parameters
        ----------
        line : str
            The line to check.
        """
        new_state = BrainsightCSVParser.State.SKIPPING
        if line.startswith("# Sample Name"):  # data header info
            new_state = BrainsightCSVParser.State.DATA_HEADER
        elif line.startswith("Sample "):
            new_state = BrainsightCSVParser.State.DATA
        elif line.startswith("# Target Name"):
            new_state = BrainsightCSVParser.State.SKIPPING
        elif line.startswith("# "):
            # we are still parsing the file header
            if self.current_state == BrainsightCSVParser.State.FILE_HEADER:
                new_state = BrainsightCSVParser.State.FILE_HEADER
            # if we surpassed the 'FILE_HEADER' state already, a line with
            # a leading '#' is a comment, which should be skipped
            else:
                new_state = BrainsightCSVParser.State.SKIPPING
        # all other lines should be skipped

        self.current_state = new_state

    def parse_line(self, line):
        """
        Parses a line and updates the corresponding attributes.

        Parameters
        ----------
        line : str
            The line to parse.
        """
        if self.current_state == BrainsightCSVParser.State.FILE_HEADER:
            self.header_file += line
        elif self.current_state == BrainsightCSVParser.State.DATA_HEADER:
            line = line[2:-1]  # strip leading comment character '# ' and trailing newline '\n'
            self.header_data = line.split('\t')
        elif self.current_state == BrainsightCSVParser.State.DATA:
            line = line[:-1]  # strip trailing newline '\n'
            data_row = line.split('\t')
            self.data.append(data_row)


def write_targets_brainsight(targets, fn_out, names=None, overwrite=True):
    """
    Writes coil position and orientations in .txt file for import into Brainsight.

    Parameters
    ----------
    targets : np.ndarray of float
    (4, 4, N_targets), Tensor containing the 4x4 matrices with coil orientation and position.
    fn_out : str
        Filename of output file.
    names : list of str, optional
        Target names (if None they will be numbered by their order).
    overwrite : bool, default: True
        Overwrite existing .txt file.

    Returns
    -------
    <file> .txt file containing the targets for import into Brainsight
    """

    targets = np.atleast_3d(targets)

    if names is None:
        names = [f"{i:04}" for i in range(targets.shape[2])]
    if isinstance(names, str):
        names = [names]
    assert targets.shape[:2] == (4, 4), 'Expecting array with shape (4, 4, N instrument marker).'

    if not fn_out.lower().endswith('.txt'):
        fn_out += '.txt'

    assert not os.path.exists(fn_out) or overwrite, '.txt file already exists. Remove or set overwrite=True.'

    with io.open(fn_out, 'w', newline='\n') as f:  # correct windows style would be \r\n, but Localite uses \n
        f.write('# Version: 12\n')
        f.write('# Coordinate system: NIfTI:Aligned\n')
        f.write('# Created by: pynibs\n')
        f.write('# Units: millimetres, degrees, milliseconds, and microvolts\n')
        f.write('# Encoding: UTF-8\n')
        f.write('# Notes: Each column is delimited by a tab. Each value within a column is delimited by a semicolon.\n')
        f.write(
                '# Target Name	Loc. X	Loc. Y	Loc. Z	m0n0	m0n1	m0n2	m1n0	m1n1	m1n2	m2n0	m2n1	m2n2\n')

        for i_t in range(targets.shape[2]):
            f.write(f'{names[i_t]}\t' +
                    f'{targets[0, 3, i_t]:.4f}\t{targets[1, 3, i_t]:.4f}\t{targets[2, 3, i_t]:.4f}\t' +
                    f'{targets[0, 0, i_t]:.4f}\t{targets[1, 0, i_t]:.4f}\t{targets[2, 0, i_t]:.4f}\t' +
                    f'{targets[0, 1, i_t]:.4f}\t{targets[1, 1, i_t]:.4f}\t{targets[2, 1, i_t]:.4f}\t' +
                    f'{targets[0, 2, i_t]:.4f}\t{targets[1, 2, i_t]:.4f}\t{targets[2, 2, i_t]:.4f}\n')


def read_targets_brainsight(fn):
    """
    Reads target coil position and orientations from .txt file and returns it as 4 x 4 x N_targets numpy array.

    Parameters
    ----------
    fn : str
        Filename of output file.

    Returns
    -------
    m_nnav : np.ndarray of float
        (4, 4, N_targets) Tensor containing the 4x4 matrices with coil orientation and position.
    """
    warnings.warn("This function is deprecated. Please use pynibs.brainsight.get_marker(markertype='Targets') instead.",
                  DeprecationWarning)
    with io.open(fn, 'r') as f:
        lines = f.readlines()
    start_idx = [i + 1 for i, l in enumerate(lines) if l.startswith("# Target")][0]
    stop_idx = [i for i, l in enumerate(lines) if l.startswith("# Sample Name") or l.startswith("# Session Name")]

    if not stop_idx:
        stop_idx = len(lines)
    else:
        stop_idx = np.min(stop_idx)

    n_targets = stop_idx - start_idx
    names = ['' for _ in range(n_targets)]
    m_nnav = np.zeros((4, 4, n_targets))
    for i_loc, i_glo in enumerate(range(start_idx, stop_idx)):
        line = lines[i_glo].split(sep='\t')
        names[i_loc] = line[0]
        m_nnav[:3, 0, i_loc] = np.array(line[4:7]).astype(float)
        m_nnav[:3, 1, i_loc] = np.array(line[7:10]).astype(float)
        m_nnav[:3, 2, i_loc] = np.array(line[10:13]).astype(float)
        m_nnav[:3, 3, i_loc] = np.array(line[1:4]).astype(float)
        m_nnav[3, 3, i_loc] = 1
    return m_nnav


def create_merged_nnav_emg_file_brainsight(fn_brainsight_nnav_info, fn_emg_info, fn_out, p2p_window_start=18,
                                           p2p_window_end=35):
    """
    Creates a merged file containing TMS coil positions and EMG data from Brainsight information.

    Parameters
    ----------
    fn_brainsight_nnav_info : str
        File path of the Brainsight nnav coil information.
    fn_emg_info : str
        File path of the EMG information.
    fn_out : str
        File path of the output merged file.
    p2p_window_start : int, default: 18
        Start of the time frame after TMS pulse where p2p value is evaluated (in ms).
    p2p_window_end : int, default: 35
        End of the time frame after TMS pulse where p2p value is evaluated (in ms).
    """
    # read EMG information
    _, ext = os.path.splitext(fn_emg_info)

    if ext == ".mat":
        try:
            with h5py.File(fn_emg_info, 'r') as data_h5:
                keys = list(data_h5.keys())
                keys = [k for k in keys if not k.startswith('#')]

                if len(keys) != 1:
                    raise ValueError(f"Key ambiguity! Please provide only exactly one data field in file: "
                                     f"{fn_emg_info}")

                emg_data = data_h5[f"/{keys[0]}/values"][:]
                emg_start = data_h5[f"/{keys[0]}/start"][0][0]
                emg_res = data_h5[f"/{keys[0]}/interval"][0][0]

                if emg_res < .1:  # time step between samples might be stored in seconds
                    emg_res *= 1e3

                num_stim = emg_data.shape[0]
                num_channels = emg_data.shape[1]
                num_emg_samples = emg_data.shape[2]

                emg_timestamps_h5_ref = data_h5[f"/{keys[0]}/frameinfo/start"][:]
                emg_timestamps = []
                for stim_idx in range(num_stim):
                    emg_timestamps.append(data_h5[emg_timestamps_h5_ref[0][stim_idx]][0][0])
                emg_timestamps = np.asarray(emg_timestamps, dtype=np.float32)
                emg_timestamps -= emg_timestamps[0]
        except OSError:
            raise NotImplementedError(f"Could not read {fn_emg_info} -- is this a matlab v7.3 .mat file?")

    elif ext == ".cfs":
        emg_data, emg_timestamps, num_stim, num_channels, num_emg_samples, sampling_rate = \
            pynibs.read_biosig_emg_data(fn_data=fn_emg_info, include_first_trigger=True, type="cfs")
        emg_res = 1000 / sampling_rate  # [ms]
        emg_start = 0  # EMG start time not included in CSF file / always 0 for CED Signal (?)
    else:
        raise ValueError(f"[Error] create_merged_nnav_emg_file_brainsight: unsupported EMG data file extension: {ext}"
                         "Only CED Signal files (.cfs) or MATLAB v.7.3 files (.mat) are supported.")

    # read nnav coil information
    with open(fn_brainsight_nnav_info, 'r') as data_txt_f:
        csv_parser = BrainsightCSVParser()
        while True:
            line = data_txt_f.readline()
            if not line:
                break
            csv_parser.check_state_transition(line)
            csv_parser.parse_line(line)

        nnav_timestamps = []
        start_time_string = current_time_string = csv_parser.data[0][24]  # 25. column of CSv data = time stamp
        start_time = datetime.datetime.strptime(start_time_string, '%H:%M:%S.%f')
        for c_idx in range(0, num_stim):
            current_time_string = csv_parser.data[c_idx][24]  # 25. column of CSv data = time stamp
            current_time = datetime.datetime.strptime(current_time_string, '%H:%M:%S.%f')
            nnav_timestamps.append((current_time - start_time).total_seconds())

        nnav_timestamps = np.asarray(nnav_timestamps, dtype=np.float32)

    # verify timesteps
    timestamp_difference = np.abs(nnav_timestamps - emg_timestamps)
    if np.any(timestamp_difference > 1):
        print("[ WARNING ] Detected time stamp differences larger than 1 second. "
              "Please check whether data from the neuronavigation are aligned with the EMG data.")

    # append EMG info to the nnav coil info
    # 1) header
    header_out = csv_parser.header_data.copy()
    header_out.append("EMG Start")
    header_out.append("EMG End")
    header_out.append("EMG Res.")
    header_out.append("EMG Channels")
    header_out.append("EMG Window Start")
    header_out.append("EMG Window End")
    for c_idx in range(num_channels):
        header_out.append(f"EMG Peak-to-peak {c_idx + 1}")
        header_out.append(f"EMG Data {c_idx + 1}")

    # 2) data
    data_text_out = []
    for stim_idx in range(len(emg_data)):
        # take what we already have in the nnav data
        row = csv_parser.data[stim_idx].copy()
        # extend it by the EMG data
        row.extend([str(emg_start),  # EMG Start
                    "(null)",  # EMG End
                    str(emg_res),  # EMG Res.
                    str(num_channels),  # EMG Channels
                    str(p2p_window_start),  # EMG Window Start
                    str(p2p_window_end)])  # EMG Window End
        for c_idx in range(num_channels):
            row.append("0")  # EMG Peak-to-peak c_idx+1
            emg_data_list = emg_data[stim_idx, c_idx, :].tolist()
            row.append(  # EMG data
                    ';'.join(map(str, emg_data_list))
            )
        data_text_out.append('\t'.join(row))

    with open(fn_out, 'w') as outfile:
        outfile.write(csv_parser.header_file)
        outfile.write("# " + '\t'.join(header_out))
        outfile.write('\n')
        outfile.write('\n'.join(data_text_out))
        outfile.write('\n')
        outfile.write("# Session Name: END")


def merge_exp_data_brainsight(subject, exp_idx, mesh_idx, coil_outlier_corr_cond=False,
                              remove_coil_skin_distance_outlier=True, coil_distance_corr=True,
                              average_emg_data_from_same_condition=False, verbose=False, plot=False,
                              start_mep=18, end_mep=35):
    """
    Merge the TMS coil positions and the mep data into an experiment.hdf5 file.

    Parameters
    ----------
    subject : pynibs.subject.Subject
        Subject object.
    exp_idx : str
        Experiment ID.
    mesh_idx : str
        Mesh ID.
    coil_outlier_corr_cond : bool, default: False
        Correct outlier of coil position and orientation (+-2 mm, +-3 deg) in case of conditions.
    remove_coil_skin_distance_outlier : bool, default: True
        Remove outlier of coil position lying too far away from the skin surface (-5/+2 mm).
    coil_distance_corr : bool, default: True
        Perform coil <-> head distance correction (coil is moved towards head surface until coil touches scalp).
    average_emg_data_from_same_condition : bool, default: False
        Flag indicating whether to average EMG data from the same condition. Only meaningful in conjunction with
        'coil_distance_correction', which averages the coil position within a condition.
    verbose : bool, default: False
        Plot output messages.
    plot : bool, default: False
        Plot MEPs and p2p evaluation.
    start_mep : int, default: 18
        Start of time frame after TMS pulse where p2p value is evaluated (in ms).
    end_mep : int, default: 35
        End of time frame after TMS pulse where p2p value is evaluated (in ms).
    """
    nii_exp_path_lst = subject.exp[exp_idx]['fn_mri_nii']
    # nii_conform_path = os.path.join(os.path.split(subject.mesh[mesh_idx]["fn_mesh_hdf5"])[0],
    #                                subject.id + "_T1fs_conform.nii.gz")
    nii_conform_path = subject.exp[exp_idx]["fn_mri_nii"][0][0]
    fn_exp_hdf5 = subject.exp[exp_idx]['fn_exp_hdf5'][0]
    fn_coil = subject.exp[exp_idx]['fn_coil'][0][0]
    fn_mesh_hdf5 = subject.mesh[mesh_idx]['fn_mesh_hdf5']
    if isinstance(subject.exp[exp_idx]['fn_exp_hdf5'][0], list):
        temp_dir = os.path.join(os.path.split(subject.exp[exp_idx]['fn_exp_hdf5'][0])[0],
                                "nnav2simnibs",
                                f"mesh_{mesh_idx}")
    else:
        temp_dir = os.path.join(os.path.split(subject.exp[exp_idx]['fn_exp_hdf5'])[0],
                                "nnav2simnibs",
                                f"mesh_{mesh_idx}")
    fn_data = None
    if "fn_data" in subject.exp[exp_idx].keys():
        if isinstance(subject.exp[exp_idx]["fn_data"], list):
            if isinstance(subject.exp[exp_idx]["fn_data"][0], str):
                fn_data = subject.exp[exp_idx]["fn_data"][0]
            else:
                fn_data = subject.exp[exp_idx]["fn_data"][0][0]
        elif isinstance(subject.exp[exp_idx]["fn_data"], str):
            fn_data = subject.exp[exp_idx]["fn_data"]
        else:
            raise ValueError(f"Invalid exp[{exp_idx}]['fn_data']: "
                             "either specify the path as a single string or as a nested list of strings [['path']]!"
                             f"Will attempt to read EMG data from exp[{exp_idx}]['fn_tms_nav']")
    else:
        print(f"No EMG data specified in exp[{exp_idx}]['fn_data']: "
              f"Will read EMG data from exp[{exp_idx}]['fn_tms_nav']")

    if "fn_tms_nav" in subject.exp[exp_idx]:
        if isinstance(subject.exp[exp_idx]["fn_tms_nav"], list):
            if isinstance(subject.exp[exp_idx]["fn_tms_nav"][0], str):
                fn_tms_nav = subject.exp[exp_idx]["fn_tms_nav"][0]
            else:
                fn_tms_nav = subject.exp[exp_idx]["fn_tms_nav"][0][0]
        elif isinstance(subject.exp[exp_idx]["fn_tms_nav"], str):
            fn_tms_nav = subject.exp[exp_idx]["fn_tms_nav"][0][0]
        else:
            raise ValueError("Invalid exp[*]['fn_tms_nav']: "
                             "Either specify the path as a single string or as a nested list of strings [['path']]!")
    else:
        raise KeyError("No exp[*]['fn_tms_nav'] found in subject object. Cannot continue to merge data,")

    if fn_data is not None:
        fn_out, out_ext = os.path.splitext(fn_tms_nav)
        fn_out += "_merged" + out_ext

        if not os.path.exists(fn_out):
            create_merged_nnav_emg_file_brainsight(
                    fn_brainsight_nnav_info=fn_tms_nav,
                    fn_emg_info=fn_data,
                    fn_out=fn_out
            )

        fn_tms_nav = fn_out

    # read Brainsight data
    ######################
    if verbose:
        print(f"Reading Brainsight data from file: {fn_tms_nav}")

    d_bs, m_nnav, n_stims = read_samples_brainsight(fn_tms_nav)

    # transform from brainsight to simnibs space
    ############################################
    if verbose:
        print(f"Transforming coil positions from Brainsight to SimNIBS space")
    print(nii_conform_path)

    m_simnibs = pynibs.nnav2simnibs(fn_exp_nii=nii_exp_path_lst[0][0],
                                    fn_conform_nii=nii_conform_path,
                                    m_nnav=m_nnav,
                                    nnav_system="brainsight",
                                    mesh_approach="headreco",
                                    fiducials=None,
                                    orientation='RAS',
                                    fsl_cmd=None,
                                    target='simnibs',
                                    temp_dir=temp_dir,
                                    rem_tmp=True,
                                    verbose=verbose)

    # create dictionary containing stimulation and physiological data
    if verbose:
        print(f"Creating dictionary containing stimulation and physiological data")

    current_scaling_factor = 1.43 if "current_scaling_factor" not in subject.exp[exp_idx].keys() \
        else subject.exp[exp_idx]["current_scaling_factor"]

    # create output directory (which is going to be persisted to experiment.hdf5)
    d = dict()
    d['coil_0'] = []
    d['coil_1'] = []
    d['coil_mean'] = []
    d['number'] = []
    d['condition'] = []
    d['current'] = []
    d['date'] = []
    d['time'] = []
    d['coil_sn'] = []
    d['patient_id'] = []

    for i in range(n_stims):
        d['coil_0'].append(m_simnibs[:, :, i])
        d['coil_1'].append(np.zeros((4, 4)) * np.NaN)
        d['coil_mean'].append(np.nanmean(np.stack((d['coil_0'][-1],
                                                   d['coil_1'][-1]), axis=2), axis=2))
        d['number'].append(d_bs['Index'][i])
        # Sample Name is unique for each pulse, so no conditions will be detected
        # d['condition'].append(d_bs['Sample Name'][i])
        # Assoc. Target will assignt he same number to pulses that had the same target/marker
        # (despite minor deviations in the true location of the coil after the pulse)
        cond = d_bs['Assoc. Target'][i]
        if cond == '(null)':
            cond = i
        d['condition'].append(cond)
        d['current'].append(1 * current_scaling_factor)
        d['date'].append(d_bs["Date"][i])
        d['time'].append(d_bs["Time"][i])
        d['coil_sn'].append(os.path.split(fn_coil)[1])
        d['patient_id'].append(subject.id)

        # add remaining data to keys that have not already been added or that should not not be added
        for key in d_bs.keys():
            if key not in ['Sample Name', 'Session Name', 'Index',
                           'Loc. X', 'Loc. Y', 'Loc. Z',
                           'm0n0', 'm0n1', 'm0n2',
                           'm1n0', 'm1n1', 'm1n2',
                           'm2n0', 'm2n1', 'm2n2',
                           'Date', 'Time'] and \
                    not (key.startswith('EMG Peak-to-peak') or
                         key.startswith('EMG Data')):

                try:
                    d[key].append(d_bs[key][i])
                except KeyError:
                    d[key] = []
                    d[key].append(d_bs[key][i])

    # add physiological raw data
    channels = subject.exp[exp_idx]["channels"]

    if verbose:
        print(f"Postprocessing MEP data")
        if plot:
            print(" Creating MEP plots ...")

    for c_idx, chan_name in enumerate(channels):
        # brainsight indexes channels 1-based
        c_idx += 1

        d[f"mep_raw_data_time_{chan_name}"] = []
        d[f"mep_filt_data_time_{chan_name}"] = []
        d[f"mep_raw_data_{chan_name}"] = []
        d[f"mep_filt_data_{chan_name}"] = []
        d[f"p2p_brainsight_{chan_name}"] = []
        d[f"p2p_{chan_name}"] = []
        d[f"mep_latency_{chan_name}"] = []

        for i in range(n_stims):
            fn_plot = None
            if plot:
                fn_channel = os.path.join(os.path.dirname(fn_tms_nav), "plots", str(chan_name))
                fn_plot = os.path.join(fn_channel, f"mep_{i:04}")
                os.makedirs(fn_channel, exist_ok=True)

            # filter data and calculate p2p values
            p2p, mep_filt_data, latency = pynibs.calc_p2p(sweep=d_bs[f"EMG Data {c_idx}"][i],
                                                          tms_pulse_time=d_bs[f"Offset"][i],
                                                          sampling_rate=1000 / d_bs["EMG Res."][i],
                                                          start_mep=start_mep, end_mep=end_mep,
                                                          measurement_start_time=float(d["EMG Start"][i]),
                                                          fn_plot=fn_plot)

            d[f"mep_raw_data_time_{chan_name}"].append(np.arange(d_bs["EMG Start"][i],
                                                                 d_bs["EMG End"][i], d_bs["EMG Res."][i]))
            d[f"mep_filt_data_time_{chan_name}"].append(
                    np.arange(d_bs["EMG Start"][i], d_bs["EMG End"][i], d_bs["EMG Res."][i]))
            d[f"mep_raw_data_{chan_name}"].append(d_bs[f"EMG Data {c_idx}"][i])
            d[f"mep_filt_data_{chan_name}"].append(mep_filt_data)
            d[f"p2p_brainsight_{chan_name}"].append(d_bs[f"EMG Peak-to-peak {c_idx}"][i])
            d[f"p2p_{chan_name}"].append(p2p)
            d[f"mep_latency_{chan_name}"].append(latency)

    # set filename of experiment.hdf5
    if subject.exp[exp_idx]["fn_exp_hdf5"] is not None or subject.exp[exp_idx]["fn_exp_hdf5"] != []:
        fn_exp_hdf5 = subject.exp[exp_idx]["fn_exp_hdf5"][0] if isinstance(subject.exp[exp_idx]["fn_exp_hdf5"], list) \
            else subject.exp[exp_idx]["fn_exp_hdf5"]

    elif subject.exp[exp_idx]["fn_exp_csv"] is not None or subject.exp[exp_idx]["fn_exp_csv"] != []:
        fn_exp_hdf5 = subject.exp[exp_idx]["fn_exp_csv"][0]

    elif fn_exp_hdf5 is None or fn_exp_hdf5 == []:
        fn_exp_hdf5 = os.path.join(subject.subject_folder, "exp", exp_idx, "experiment.hdf5")

    # remove coil position outliers (in case of conditions)
    if coil_outlier_corr_cond:
        if verbose:
            print("Removing coil position outliers")
        d = pynibs.coil_outlier_correction_cond(exp=d, outlier_angle=5., outlier_loc=3., fn_exp_out=fn_exp_hdf5)

    # perform coil <-> head distance correction
    if coil_distance_corr:
        if verbose:
            print("Performing coil <-> head distance correction")
        d = pynibs.coil_distance_correction(exp=d,
                                            fn_geo_hdf5=fn_mesh_hdf5,
                                            remove_coil_skin_distance_outlier=remove_coil_skin_distance_outlier,
                                            fn_plot=os.path.split(fn_exp_hdf5)[0])

    d_avg = dict()
    for key in d.keys():
        d_avg[key] = []

    # TODO: This only meaningfully works in conjunction with 'coil_distance_correction', which averages the
    #       coil position within a condition. For each condition, we wil then have 1) and average coil position,
    #       2) an average EMG sample, 3) only 1 full dataset (all but one samples are discarded after averaging)
    if average_emg_data_from_same_condition:
        if verbose:
            print("Averaging samples from the same condition")

        conditions_per_sample = np.array(d['condition'])
        conditions_unique = np.unique(conditions_per_sample)

        # Create new, dedicated dictionary with all EMG data to:
        # 1) collect all EG related keys of the main dictionary
        # 2) convert the stored lists to numpy arrays (for list-based indexing)
        all_emg_data = dict()
        for c_idx, chan_name in enumerate(channels):
            all_emg_data[f"mep_raw_data_{chan_name}"] = np.array(d[f"mep_raw_data_{chan_name}"])
            all_emg_data[f"mep_filt_data_{chan_name}"] = np.array(d[f"mep_filt_data_{chan_name}"])
            all_emg_data[f"p2p_brainsight_{chan_name}"] = np.array(d[f"p2p_brainsight_{chan_name}"])
            all_emg_data[f"p2p_{chan_name}"] = np.array(d[f"p2p_{chan_name}"])

        # create a single entry in the new result dictionary (d_avg) for each condition
        for condition_id in conditions_unique:
            condition_idcs = np.where(conditions_per_sample == condition_id)[0]

            # just copy the first occurrence (within a condition) of non-emg related keys
            if len(condition_idcs) > 0:
                for key in d.keys():
                    if key not in all_emg_data.keys():
                        d_avg[key].append(d[key][condition_idcs[0]])

                # average all emg-related keys of a condition
                for key in all_emg_data.keys():
                    d_avg[key].append(np.mean(all_emg_data[key][condition_idcs], axis=0))

        d = d_avg

    # create dictionary of stimulation data
    d_stim_data = dict()
    d_stim_data["coil_0"] = d["coil_0"]
    d_stim_data["coil_1"] = d["coil_1"]
    d_stim_data["coil_mean"] = d["coil_mean"]
    d_stim_data["number"] = d["number"]
    d_stim_data["condition"] = d["condition"]
    d_stim_data["current"] = d["current"]
    d_stim_data["date"] = d["date"]
    d_stim_data["time"] = d["time"]
    d_stim_data["Creation Cause"] = d["Creation Cause"]
    d_stim_data["Offset"] = d["Offset"]

    # create dictionary of raw physiological data
    d_phys_data_raw = dict()
    d_phys_data_raw["EMG Start"] = d["EMG Start"]
    d_phys_data_raw["EMG End"] = d["EMG End"]
    d_phys_data_raw["EMG Res."] = d["EMG Res."]
    d_phys_data_raw["EMG Channels"] = d["EMG Channels"]
    d_phys_data_raw["EMG Window Start"] = d["EMG Window Start"]
    d_phys_data_raw["EMG Window End"] = d["EMG Window End"]

    for chan in channels:
        d_phys_data_raw[f"mep_raw_data_time_{chan}"] = d[f"mep_raw_data_time_{chan}"]
        d_phys_data_raw[f"mep_raw_data_{chan}"] = d[f"mep_raw_data_{chan}"]

    # create dictionary of postprocessed physiological data
    d_phys_data_postproc = dict()

    for chan in channels:
        d_phys_data_postproc[f"mep_filt_data_time_{chan}"] = d[f"mep_filt_data_time_{chan}"]
        d_phys_data_postproc[f"mep_filt_data_{chan}"] = d[f"mep_filt_data_{chan}"]
        d_phys_data_postproc[f"p2p_brainsight_{chan}"] = d[f"p2p_brainsight_{chan}"]
        d_phys_data_postproc[f"p2p_{chan}"] = d[f"p2p_{chan}"]
        d_phys_data_postproc[f"mep_latency_{chan}"] = d[f"mep_latency_{chan}"]

    # create pandas dataframes from dicts
    df_stim_data = pd.DataFrame.from_dict(d_stim_data)
    df_phys_data_raw = pd.DataFrame.from_dict(d_phys_data_raw)
    df_phys_data_postproc = pd.DataFrame.from_dict(d_phys_data_postproc)

    # save in .hdf5 file
    if verbose:
        print(f"Saving experimental data to file: {fn_exp_hdf5}")
    df_stim_data.to_hdf(fn_exp_hdf5, "stim_data")
    df_phys_data_raw.to_hdf(fn_exp_hdf5, "phys_data/raw/EMG")
    df_phys_data_postproc.to_hdf(fn_exp_hdf5, "phys_data/postproc/EMG")


def read_samples_brainsight(fn_tms_nav):
    """
    Reads sample data from a Brainsight navigation system file.

    Parameters
    ----------
    fn_tms_nav : str
        The file path of the Brainsight navigation system file.

    Returns
    -------
    d_bs : collections.OrderedDict
        An ordered dictionary containing the parsed data from the Brainsight file.
        The keys of the dictionary correspond to the column headers in the Brainsight file,
        and the values are lists of the data in those columns.
    m_nnav : numpy.ndarray
        A 3D numpy array of shape (4, 4, n_stims) containing the 4x4 transformation matrices for each stimulation.
    n_stims : int
        The number of stimulations, inferred from the number of rows in the Brainsight file.
    """
    warnings.warn("This function is deprecated. Please use pynibs.brainsight.get_marker(markertype='Samples') instead.",
                  DeprecationWarning)
    d_bs = OrderedDict()
    with io.open(fn_tms_nav, 'r') as f:
        lines = f.readlines()
    start_idx = [i + 1 for i, l in enumerate(lines) if l.startswith("# Sample Name")][0]
    stop_idx = [i for i, l in enumerate(lines) if l.startswith("# Session Name")]
    if not stop_idx:
        stop_idx = len(lines)
    else:
        stop_idx = np.min(stop_idx)
    n_stims = stop_idx - start_idx
    m_nnav = np.zeros((4, 4, n_stims))
    keys = lines[start_idx - 1].split('\t')
    keys[0] = keys[0].replace('# ', '')
    keys[-1] = keys[-1].replace('\n', '')
    # create brainsight dict
    for key in keys:
        d_bs[key] = []
    # collect data; for each line...
    for idx_coil_sample, idx_line in enumerate(range(start_idx, stop_idx)):
        line = lines[idx_line].split(sep='\t')

        # ...populate the keys of the brainsight-dict
        for idx_key, key in enumerate(d_bs.keys()):
            # standard (string) entries can be appended straightforwardly
            if key in ["Sample Name", "Session Name", "Creation Cause", "Crosshairs Driver", "Date", "Time",
                       "EMG Channels", "Assoc. Target"]:
                d_bs[key].append(line[idx_key])
            # other keys include:  "Loc. X/Y/Z", "EMG Data 1/2", "EMG Peak-to-Peak 1" etc...
            else:
                # each sample of EMG data is separated by a semicolon
                if ";" in line[idx_key]:
                    d_bs[key].append(np.array(line[idx_key].split(';')).astype(float))
                # other non-EMG-data (numeric) entries
                else:
                    # if key was selected for export but no corresponding data could be collected, "(null)" is entered
                    if line[idx_key] != "(null)":
                        try:
                            d_bs[key].append(float(line[idx_key]))
                        except Exception:
                            pass
                    else:
                        d_bs[key].append(float(-1))

            # assemble coil configuration matrix
            if key in ["Loc. X"]:
                m_nnav[0, 3, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["Loc. Y"]:
                m_nnav[1, 3, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["Loc. Z"]:
                m_nnav[2, 3, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m0n0"]:
                m_nnav[0, 0, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m0n1"]:
                m_nnav[1, 0, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m0n2"]:
                m_nnav[2, 0, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m1n0"]:
                m_nnav[0, 1, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m1n1"]:
                m_nnav[1, 1, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m1n2"]:
                m_nnav[2, 1, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m2n0"]:
                m_nnav[0, 2, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m2n1"]:
                m_nnav[1, 2, idx_coil_sample] = float(d_bs[key][-1])
            elif key in ["m2n2"]:
                m_nnav[2, 2, idx_coil_sample] = float(d_bs[key][-1])

            m_nnav[3, 3, idx_coil_sample] = 1
    return d_bs, m_nnav, n_stims


def transform_brainsight(data, col_names):
    """
    Transforms Brainsight coil position/orientation into SimNIBS matsimnibs

    data: list of lists with positions
    col_names: list of str

    Returns:
    --------
    names : list of str
        Samples/target names. len(names) = n_pos
    matsimnibs: np.ndarray
        Coil position/orientation in matsimnibs style with Brainsight coil axes definition
        shape: (n_pos, 4, 4)

    Written by Ole Numssen, numssen@cbs.mpg.de, 2022.
    """
    matsimnibs = np.zeros((len(data), 4, 4))
    pos_names = []
    for pos, i in zip(data, range(len(data))):
        p_dict = {x: pos[k] for x, k in zip(col_names, range(len(col_names)))}

        m = [[p_dict['m0n0'], p_dict['m1n0'], p_dict['m2n0'], p_dict['Loc. X']],
             [p_dict['m0n1'], p_dict['m1n1'], p_dict['m2n1'], p_dict['Loc. Y']],
             [p_dict['m0n2'], p_dict['m1n2'], p_dict['m2n2'], p_dict['Loc. Z']],
             [0, 0, 0, 1]]

        matsimnibs[i] = np.array(m).astype(float)
        pos_names.append(pos[0])

    # adjust coil axes definition to simnibs style
    return pos_names, matsimnibs  # @ self.simnibs2brainsight()


def simnibs2brainsight():
    """
    Flip matrix for brainsight -> simnibs and vice versa.

    Example:
        if brainsight_mat.shape = (n_pos, 4, 4):
        simnibs_mat = brainsight_mat @ simnibs2localite()
        localite_mat = simnibs_mat @ brainsight2localite()
    """
    return np.array([
        [-1, +0, +0, +0],  # +x -> -x
        [+0, +1, +0, +0],  # +y -> +y
        [+0, +0, -1, +0],  # +z -> -z
        [0, 0, 0, 1]])


def get_marker(fn, markertype):
    """
    Read instrument-marker and conditions from Brainsight session.

    BrainSight stores recorded coil placements as 'Samples' and predefined coil placements
    as 'Targets'.

    Coil axes definitions as defined by Brainsight.

    Parameters
    ----------
    fn : str
        Path to instrument-marker-file.
    markertype : str
        One of ['Samples', 'Targets', 'TriggerMarker', 'InstrumentMarker'].

    Returns
    -------
    matsimnibs : np.ndarray of float
        (M, 4, 4) Instrument marker matrices.
    marker_descriptions : list of str
        Labels of the marker conditions.
    marker_times : list of str
        YYYY-MM-DD HH:MM:SS.sss time stamps of the samples. [-1] for InstrumentMarkers.
    """
    assert markertype in ['Samples', 'Targets', 'TriggerMarker', 'InstrumentMarker']

    # init empty values in case nothing is found in fn
    coord_sys, encoding, version = '', '', -1
    data_targets, data_samples = [], []

    with open(fn, 'r') as f:
        # read header
        while True:
            line = f.readline().rstrip()
            if line.startswith('# Target Name') or line.startswith('# Sample Name'):
                break
            elif line.startswith('# Version:'):
                version = int(line.replace("# Version: ", ""))

        if version == -1:
            raise ValueError(f"Cannot read version from {fn}")

        # read data, probably 'Targets'
        if line.startswith('# Target Name'):
            # get column names
            col_names_targets = line.replace('# ', '').split('\t')

            # read all target lines
            while line:
                line = f.readline()
                if line.startswith('#'):
                    break
                if line:
                    data_targets.append(line.rstrip().split('\t'))

        if line.startswith('# Sample Name'):
            # get column names
            col_names_samples = line.replace('# ', '').split('\t')

            # read all sample lines
            while line:
                line = f.readline()
                if line.startswith('#'):
                    break
                data_samples.append(line.rstrip().split('\t'))

        # get matsimnibs arrays in simnibs space and axes definition
        if len(data_targets):
            names_targets, matsimnibs_targets = transform_brainsight(data_targets, col_names_targets)
        else:
            names_targets, matsimnibs_targets = [], None

        if len(data_samples):
            names_samples, matsimnibs_samples = transform_brainsight(data_samples, col_names_samples)
        else:
            names_samples, matsimnibs_samples = [], None

        if (matsimnibs_targets is None or matsimnibs_targets.size == 0) and \
                (matsimnibs_samples is None or matsimnibs_samples.size == 0):
            raise ValueError(f"Could not find any targets in {fn}.")

        if markertype in ['Targets', 'InstrumentMarker']:
            return matsimnibs_targets, names_targets, [-1] * len(names_targets)
        else:
            try:
                # get timings for samples
                date_col = np.where(np.array(col_names_samples) == "Date")[0][0]
                time_col = np.where(np.array(col_names_samples) == "Time")[0][0]

                times_samples = [f"{sample[date_col]} {sample[time_col]}" for sample in data_samples]
            except Exception:
                print(f"Cannot extract date and time from {fn}.")
                times_samples = [-1] * len(names_samples)

            return matsimnibs_samples, names_samples, times_samples


def read_triggermarker(fn):
    """
    Read instrument-marker and conditions from Brainsight session.

    BrainSight stores recorded coil placements as 'Samples' and predefined coil placements
     as 'Targets'.

    Coil axes definitions as defined by Brainsight.
    dI/dt is returned as list of [-1] to keep consistent with read_triggermarker functions for other neuronavigation
     systems.

    Parameters
    ----------
    fn : str or list of str
        Path to instrument-marker-file.

    Returns
    -------
    matsimnibs : np.ndarray of float
        (4, 4, N) Instrument marker matrices
    didt : np.ndarray of float
        (N) Rate of change of coil current in (A/us).
    stim_int : np.ndarray of float
        (N) Stimulator intensity in (% MSO).
    marker_descriptions : list of str
        Labels of the marker conditions.
    marker_times : list of str
        YYYY-MM-DD HH:MM:SS.sss time stamps of the samples. [-1] for InstrumentMarkers.
    """
    # init empty values in case nothing is found in fn
    coord_sys, encoding, version = '', '', -1
    data_targets, data_samples = [], []

    with open(fn, 'r') as f:
        # read header
        while True:
            line = f.readline().rstrip()
            if line.startswith('# Target Name') or line.startswith('# Sample Name'):
                break
            elif line.startswith('# Version:'):
                version = int(line.replace("# Version: ", ""))

        if version == -1:
            raise ValueError(f"Cannot read version from {fn}")

        # read data, probably 'Targets'
        if line.startswith('# Target Name'):
            # read all target lines
            while line:
                line = f.readline()
                if line.startswith('#'):
                    break

        if line.startswith('# Sample Name'):
            # get column names
            col_names_samples = line.replace('# ', '').split('\t')

            # read all sample lines
            while line:
                line = f.readline()
                if line.startswith('#'):
                    break
                data_samples.append(line.rstrip().split('\t'))

        if len(data_samples):
            names_samples, matsimnibs_samples = transform_brainsight(data_samples, col_names_samples)
        else:
            names_samples, matsimnibs_samples = [], None

        if (matsimnibs_samples is None or matsimnibs_samples.size == 0):
            raise ValueError(f"Could not find any targets in {fn}.")

        # get timings for samples
        date_col = np.where(np.array(col_names_samples) == "Date")[0][0]
        time_col = np.where(np.array(col_names_samples) == "Time")[0][0]
        times_samples = [f"{sample[date_col]} {sample[time_col]}" for sample in data_samples]

        intensity_col = np.where(np.array(col_names_samples) == "Stim. Power")[0][0]
        intensity_samples = [sample[intensity_col] for sample in data_samples]

        matsimnibs_samples = np.moveaxis(matsimnibs_samples, 0, -1)

        return matsimnibs_samples, [-1] * len(names_samples), intensity_samples, names_samples, times_samples



import os
import copy
import time
import h5py
import inspect
import warnings
import itertools
import subprocess
import numpy as np
from scipy.special import binom
from sklearn.neighbors import KernelDensity
import pynibs


def tal2mni(coords, direction='tal2mni', style='nonlinear'):
    """
    Transform Talairach coordinates into (SPM) MNI space and vice versa.

    This is taken from https://imaging.mrc-cbu.cam.ac.uk/imaging/MniTalairach and
    http://gibms.mc.ntu.edu.tw/bmlab/tools/data-analysis-codes/mni2tal-m/

    Parameters
    ----------
    coords : np.ndarray or list
        x,y,z coordinates.
    direction : str, default: 'tal2mni
        Transformation direction. One of ('tal2mni', 'mni2tal').
    style : str, default: 'nonlinear'
        Transformation style. One of ('linear', 'nonlinear').

    Returns
    -------
    coords_trans : np.ndarray

    """
    assert direction in ['tal2mni',
                         'mni2tal'], f"direction parameter '{direction}' invalid. Choose 'tal2mni' or 'mni2tal'."
    assert style in ['linear', 'nonlinear'], f"style parameter '{style}' invalid. Choose 'linear' or 'nonlinear'."
    if len(coords) == 3:
        coords = np.hstack((coords, 1))
    if style == 'linear':
        mat = np.array([[0.88, 0, 0, -0.8], [0, 0.97, 0, -3.32], [0, 0.05, 0.88, -0.44], [0, 0, 0, 1]])
        if direction == 'mni2tal':
            return np.dot(mat, coords)[:3]
        elif direction == 'tal2mni':
            return np.dot(np.linalg.inv(mat), coords)[:3]
    elif style == 'nonlinear':
        upper_mat = np.array([[0.99, 0, 0, 0],
                              [0, 0.9688, 0.0460, 0],
                              [0, -0.0485, 0.9189, 0],
                              [0, 0, 0, 1]])

        lower_mat = np.array([[0.99, 0, 0, 0],
                              [0, 0.9688, 0.042, 0],
                              [0, -0.0485, 0.839, 0],
                              [0, 0, 0, 1]])
        if direction == 'mni2tal':
            pass
        elif direction == 'tal2mni':
            upper_mat = np.linalg.inv(upper_mat)
            lower_mat = np.linalg.inv(lower_mat)
        if coords[2] > 0:
            # above AC
            return np.dot(upper_mat, coords)[:3]
        else:
            # below AC
            return np.dot(lower_mat, coords)[:3]


def rd(array, array_ref):
    """
    Determine the relative difference between input data and reference data.

    Parameters
    ----------
    array : np.ndarray
        input data [ (x), y0, y1, y2 ... ].
    array_ref : np.ndarray
        reference data [ (x_ref), y0_ref, y1_ref, y2_ref ... ]
        if array_ref is 1D, all sizes have to match.

    Returns
    -------
    rd : ndarray of float
        (array.shape[1]) Relative difference between the columns of array and array_ref.
    """

    return np.linalg.norm(array - array_ref) / np.linalg.norm(array_ref)


def generalized_extreme_value_distribution(x, mu, sigma, k):
    """
    Generalized extreme value distribution.

    Parameters
    ----------
    x : ndarray of float
        (n_x) Events.
    mu : float
        Mean value.
    sigma : float
        Standard deviation.
    k : float
        Shape parameter.

    Returns
    -------
    y : ndarray of float
        (n_x) Probability density of events
    """
    y = 1 / sigma * np.exp(-(1 + (k * (x - mu)) / sigma) ** (-1 / k)) * (1 + (k * (x - mu)) / sigma) ** (-(1 + 1 / k))

    return y


def differential_evolution(fobj, bounds, mut=0.8, crossp=0.7, popsize=20, its=1000, **kwargs):
    """
    Differential evolution optimization algorithm

    Parameters
    ----------
    fobj : function object
        Function to optimize.
    bounds : dict
        Dictionary containing the bounds of the free variables to optimize.
    mut : float, default: 0.8
        Mutation factor.
    crossp : float, default: 0.7
        Cross population factor.
    popsize : int, default: 20
        Population size.
    its : int, default: 1000
        Number of iterations.
    kwargs : dict
        Arguments passed to fobj (constants etc...).

    Returns
    -------
    best : dict
        Dictionary containing the best values.
    fitness : float
        Fitness value of best solution.
    """
    if kwargs is None:
        kwargs = dict()

    fobj_args = inspect.getfullargspec(fobj).args

    if "bounds" in fobj_args:
        kwargs["bounds"] = bounds
    params_str = list(bounds.keys())

    # set up initial simulations
    dimensions = len(bounds)
    pop = np.random.rand(popsize, dimensions)

    min_b = np.zeros(dimensions)
    max_b = np.zeros(dimensions)

    for i, key in enumerate(bounds):
        min_b[i] = bounds[key][0]
        max_b[i] = bounds[key][1]

    diff = np.fabs(min_b - max_b)
    pop_denorm = min_b + pop * diff

    print("Initial simulations:")
    print("====================")

    fitness = np.zeros(len(pop_denorm))

    for i, po in enumerate(pop_denorm):

        for i_key, key in enumerate(params_str):
            kwargs[key] = po[i_key]

        fitness[i] = fobj(**kwargs)

    best_idx = np.argmin(fitness)
    best = pop_denorm[best_idx]

    parameter_str = [f"{params_str[i_p]}={p_:.5f}" for i_p, p_ in enumerate(pop_denorm[best_idx])]
    print(f"-> Fittest: {fitness[best_idx]:.3f} / " + ", ".join(parameter_str))

    for i in range(its):
        print(f"Iteration: {i}")
        print(f"==============")

        trial_denorm = []
        trial = []

        # create new parameter sets
        for j in range(popsize):
            idxs = [idx for idx in range(popsize) if idx != j]
            a, b, c = pop[np.random.choice(idxs, 3, replace=False)]
            mutant = np.clip(a + mut * (b - c), 0, 1)
            cross_points = np.random.rand(dimensions) < crossp

            if not np.any(cross_points):
                cross_points[np.random.randint(0, dimensions)] = True

            trial.append(np.where(cross_points, mutant, pop[j]))
            trial_denorm.append(min_b + trial[j] * diff)

        # run likelihood function
        f = np.zeros(len(trial))

        for j, tr in enumerate(trial_denorm):
            for i_key, key in enumerate(params_str):
                kwargs[key] = tr[i_key]
            f[j] = fobj(**kwargs)

        for j in range(popsize):
            if f[j] < fitness[j]:
                fitness[j] = copy.deepcopy(f[j])
                pop[j] = trial[j]
                if f[j] < fitness[best_idx]:
                    best_idx = j
                    best = trial_denorm[j]

        parameter_str = [f"{params_str[i_p]}={p_:.5f}" for i_p, p_ in enumerate(best)]
        print(f"-> Fittest: {fitness[best_idx]:.3f} / " + ", ".join(parameter_str))

    best_dict = dict()
    for i_key, key in enumerate(params_str):
        best_dict[key] = best[i_key]

    return best_dict, fitness[best_idx]


def sigmoid_log_p(x, p):
    y = np.log10(p[0] / (1 + np.exp(-p[1] * (x - p[2]))))
    return y


def likelihood_posterior(x, y, fun, bounds=None, verbose=True, normalized_params=False, **params):
    """
    Determines the likelihood of the data following the function "fun" assuming a two
    variability source of the data pairs (x, y) using the posterior distribution.

    Parameters
    ----------
    x : ndarray of float
        (n_points) x data.
    y : ndarray of float
        (n_points) y data
    fun : function
        Function to fit the data to (e.g. sigmoid).
    bounds : dict, optional
        Dictionary containing the bounds of "sigma_x" and "sigma_y" and the free parameters of fun.
    verbose : bool, default: True
        Print function output after every calculation.
    normalized_params : bool, default: False
        Are the parameters passed in normalized space between [0, 1]? If so, bounds are used to
        denormalize them before calculation.
    **params : dict
        Free parameters to optimize. Contains "sigma_x", "sigma_y", and the free parameters of fun.

    Returns
    -------
    l : float
        Negative likelihood.
    """
    start = time.time()

    # read arguments from function
    # args = inspect.getfullargspec(fun).args

    # extract parameters
    sigma_x = params["sigma_x"]
    sigma_y = params["sigma_y"]

    del params["sigma_x"], params["sigma_y"]

    # denormalize parameters from [0, 1] to bounds
    if normalized_params:
        if bounds is None:
            raise ValueError("Please provide bounds if parameters were passed normalized!")
        sigma_x = sigma_x * (bounds["sigma_x"][1] - bounds["sigma_x"][0]) + bounds["sigma_x"][0]
        sigma_y = sigma_y * (bounds["sigma_y"][1] - bounds["sigma_y"][0]) + bounds["sigma_y"][0]

        for key in enumerate(params):
            params[key] = params[key] * (bounds[key][1] - bounds[key][0]) + bounds[key][0]

    if sigma_x < 0:
        sigma_x = 0

    if sigma_y < 0:
        sigma_y = 0

    # determine posterior of DVS model with test data
    x_pre = np.linspace(np.min(x), np.max(x), 500000)
    x_post = x_pre + np.random.normal(loc=0., scale=sigma_x, size=len(x_pre))
    y_post = fun(x_post, **params) + np.random.normal(loc=0., scale=sigma_y, size=len(x_pre))

    # bin data
    n_bins = 50
    dx_bins = (np.max(x_pre) - np.min(x_pre)) / n_bins
    x_bins_loc = np.linspace(np.min(x_pre) + dx_bins / 2, np.max(x_pre) - dx_bins / 2, n_bins)

    # determine probabilities of observations
    kde = KernelDensity(bandwidth=0.01, kernel='gaussian')

    likelihood = []

    for i in range(n_bins):
        mask = np.logical_and(x_pre >= (x_bins_loc[i] - dx_bins / 2), x_pre < (x_bins_loc[i] + dx_bins / 2))
        mask_data = np.logical_and(x >= (x_bins_loc[i] - dx_bins / 2), x < (x_bins_loc[i] + dx_bins / 2))

        if np.sum(mask_data) == 0:
            continue

        # determine kernel density estimate
        try:
            kde_bins = kde.fit(y_post[mask][:, np.newaxis])
        except ValueError:
            warnings.warn("kde.fit(y_post[mask][:, np.newaxis]) yield NaN ... skipping bin")
            continue

        # get probability densities at data
        kde_y_post_bins = np.exp(kde_bins.score_samples(y[mask_data][:, np.newaxis]))

        likelihood.append(kde_y_post_bins)

    likelihood = np.concatenate(likelihood)

    # mask out zero probabilities
    likelihood[likelihood == 0] = 1e-100

    # determine log likelihood
    likelihood = np.sum(np.log10(likelihood))

    stop = time.time()

    if verbose:
        parameter_str = [f"{p_}={params[p_]:.5f}" for p_ in params]
        print(f"Likelihood: {likelihood:.1f} / sigma_x={sigma_x:.5f}, sigma_y={sigma_y:.5f}, " +
              ", ".join(parameter_str) + f"({stop - start:.2f} sec)")

    return -likelihood


def mutual_coherence(array):
    """
    Calculate the mutual coherence of a matrix A. It can also be referred as the cosine of the smallest angle
    between two columns.

    mutual_coherence = mutual_coherence(array)

    Parameters
    ----------
    array : ndarray of float
        Input matrix.

    Returns
    -------
    mutual_coherence : float
        Mutual coherence.
    """

    array = array / np.linalg.norm(array, axis=0)[np.newaxis, :]
    t = np.matmul(array.conj().T, array)
    np.fill_diagonal(t, 0.0)
    mu = np.max(t)

    # s = np.sqrt(np.diag(t))
    # s_sqrt = np.diag(s)
    # mu = np.max(1.0*(t-s_sqrt)/np.outer(s, s))

    return mu


def get_cartesian_product(array_list):
    """
    Generate a cartesian product of input arrays (all combinations).

    cartesian_product = get_cartesian_product(array_list)

    Parameters
    ----------
    array_list : list of 1D ndarray of float
        Arrays to compute the cartesian product with.

    Returns
    -------
    cartesian_product : ndarray of float
        (M, len(arrays)) Array containing the cartesian products (all combinations of input vectors).

    Examples
    --------
     .. code-block:: python
        import pygpc
        out = pygpc.get_cartesian_product(([1, 2, 3], [4, 5], [6, 7]))
        out
    """
    cartesian_product = [element for element in itertools.product(*array_list)]
    return np.array(cartesian_product)


def norm_percentile(data, percentile):
    """
    Normalizes data to a given percentile.

    Parameters
    ----------
    data : np.ndarray
        (n_data, ) Dataset to normalize.
    percentile : float
        Percentile of normalization value [0 ... 100].

    Returns
    -------
    data_norm : np.ndarray
        (n_data, ) Normalized dataset.
    """
    return data / np.percentile(data, percentile)


def compute_chunks(seq, num):
    """
    Splits up a sequence _seq_ into _num_ chunks of similar size.
    If len(seq) < num, (num-len(seq)) empty chunks are returned so that len(out) == num.

    Parameters
    ----------
    seq : list of something
        (n_ele) List containing data or indices, which is divided into chunks.
    num : int
        Number of chunks to generate.

    Returns
    -------
    out : list of num sublists
        num sub-lists of seq with each of a similar number of elements (or empty).
    """
    assert len(seq) > 0
    assert num > 0
    assert isinstance(seq, list), f"{type(seq)} can't be chunked. Provide list."
    avg = len(seq) / float(num)
    n_empty = 0  # if len(seg) < num, how many empty lists to append to return?

    if avg < 1:
        # raise ValueError("seq/num ration too small: " + str(avg))
        avg = 1
        n_empty = num - len(seq)

    out = []
    last = 0.0

    while last < len(seq):
        # if only one element would be left in the last run, add it to the current
        if (int(last + avg) + 1) == len(seq):
            last_append_idx = int(last + avg) + 1
        else:
            last_append_idx = int(last + avg)

        out.append(seq[int(last):last_append_idx])

        if (int(last + avg) + 1) == len(seq):
            last += avg + 1
        else:
            last += avg

    # append empty lists if len(seq) < num
    out += [[]] * n_empty

    return out


def bash(command):
    """
    Executes bash command and returns output message in stdout (uses os.popen).

    Parameters
    ----------
    command : str
        Bash command.

    Returns
    -------
    output : str
        Output from stdout.
    error : str
        Error message from stdout.
    """
    print(("Running " + command))
    return os.popen(command).read()
    # process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    # output, error = process.communicate()
    # return output, error


def bash_call(command):
    """
    Executes bash command and returns output message in stdout (uses subprocess.Popen).

    Parameters
    ----------
    command : str
        bash command.
    """
    subprocess.Popen(command, shell=True)


def invert(trans):
    """
    Invert rotation matrix.

    Parameters
    ----------
    trans : np.ndarray of float
        (3, 3) Rotation matrix.

    Returns
    -------
    rot_inv : np.ndarray of float
        (3, 3) Inverse rotation matrix.
    """
    rot = pynibs.normalize_rot(trans[:3, :3].flatten())
    result = np.zeros((4, 4))
    result[3, 3] = 1.0
    t = -rot.T.dot(trans[:3, 3])
    result[:3, :3] = rot.T
    result[:3, 3] = t
    return result


def list2dict(l):
    """
    Transform list of dicts with same keys to dict of list

    Parameters
    ----------
    l : list of dict
        List containing dictionaries with same keys.

    Returns
    -------
    d : dict of lists
        Dictionary containing the entries in a list.
    """
    n = len(l)
    keys = l[0].keys()
    d = dict()

    for key in keys:
        d[key] = [0 for _ in range(n)]
        for i in range(n):
            d[key][i] = l[i][key]

    return d


def recursive_len(item):
    """
    Determine len of list of lists (recursively).

    Parameters
    ----------
    item : list of list
        List of list.

    Returns
    -------
    len : int
        Total length of list of lists.
    """
    if type(item) == list:
        return sum(recursive_len(subitem) for subitem in item)
    else:
        return 1


def add_center(var):
    """
    Adds center to argument list.

    Parameters
    ----------
    var : list of float
        (2) List containing two values [f1,f2].

    Returns
    -------
    out: list of float
        (3) List containing the average value in the middle [f1, mean(f1,f2), f2].
    """
    return [var[0], sum(var) / 2, var[1]]


def unique_rows(a):
    """
    Returns the unique rows of np.array(a).

    Parameters
    ----------
    a : np.ndarray of float
        (m, n) Array to search for double row entries.

    Returns
    -------
    a_unique : np.array
        (k, n) array a with only unique rows.
    """
    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    _, idx = np.unique(b, return_index=True)
    return a[idx]
    # alternative but ~10x slower:
    # surface_points=np.vstack({tuple(row) for row in surface_points})


def calc_n_network_combs(n_e, n_c, n_i):
    """
    Determine number of combinations if all conditions may be replaced between N_i elements (mixed interaction).

    Parameters
    ----------
    n_e : int
        Number of elements in the ROI.
    n_c : int
        Number of conditions (I/O curves).
    n_i : int
        Number of maximum interactions.

    Returns
    -------
    n_comb : int
        Number of combinations.
    """
    return binom(n_e, n_i) * np.sum([((n_i - 1) ** k) * binom(n_c, k) for k in range(1, n_c)])


def load_muaps(fn_muaps, fs=1e6, fs_downsample=1e5):
    # load MUAPs and downsample
    with h5py.File(fn_muaps, "r") as f:
        muaps_orig = f["MUAPShapes"][:]
    N_MU = muaps_orig.shape[1]

    t_muap_orig = np.linspace(0, 1 / fs * (muaps_orig.shape[0] - 1), muaps_orig.shape[0])
    t_muap = np.linspace(0, t_muap_orig[-1], int(t_muap_orig[-1] * fs_downsample + 1))
    muaps = np.zeros((len(t_muap), N_MU))

    for i in range(N_MU):
        muaps[:, i] = np.interp(t_muap, t_muap_orig, muaps_orig[:, i])

    return muaps, t_muap


def cross_product(A, B):
    """
    Evaluates the cross product between the vector pairs in a and b using pure Python.

    Parameters
    ----------
    A : np.ndarray of float
        (2, (N, 3))
    B : np.ndarray of float
        (2, (N, 3)) Input vectors, the cross product is evaluated between.

    Returns
    -------
    c : np.ndarray of float
        (N, 3) Cross product between vector pairs in a and b.
    """
    c1 = np.multiply(A[:, 1], B[:, 2]) - np.multiply(A[:, 2], B[:, 1])
    c2 = np.multiply(A[:, 2], B[:, 0]) - np.multiply(A[:, 0], B[:, 2])
    c3 = np.multiply(A[:, 0], B[:, 1]) - np.multiply(A[:, 1], B[:, 0])
    # C=np.array(np.multiply(A[:, 1], B[:, 2]) - np.multiply(A[:, 2], B[:, 1]),
    #            np.multiply(A[:, 2], B[:, 0]) - np.multiply(A[:, 0], B[:, 2]),
    #            np.multiply(A[:, 0], B[:, 1]) - np.multiply(A[:, 1], B[:, 0]))
    return np.vstack([c1, c2, c3]).transpose()


def cross_product_einsum2(a, b):
    """
    Evaluates the cross product between the vector pairs in a and b using the double Einstein sum.

    Parameters
    ----------
    a : np.ndarray of float
        (2, (N, 3))
    b : np.ndarray of float
        (2, (N, 3)) Input vectors, the cross product is evaluated between.

    Returns
    -------
    c : np.ndarray of float
        (N, 3) Cross product between vector pairs in a and b.
    """
    eijk = np.zeros((3, 3, 3))
    eijk[0, 1, 2] = eijk[1, 2, 0] = eijk[2, 0, 1] = 1
    eijk[0, 2, 1] = eijk[2, 1, 0] = eijk[1, 0, 2] = -1

    return np.einsum('iak,ak->ai', np.einsum('ijk,aj->iak', eijk, a), b)


def intersection_vec_plan(ray_dir, ray_origin, plane_n, plane_p, eps=1e-6):
    """
    Computes intersection between vector ('ray') and plane.

    Parameters
    ----------
    ray_dir : np.ndarray
        (3,) (Rotated) vector direction.
    ray_origin : np.ndarray
        (3,) Any point on the ray.
    plane_n : np.ndarray
        (3,) Plane normal.
    plane_p : np.ndarray
        (3,) Any point on the plane.
    eps : float, default=1e-6
        Resolution.

    Returns
    -------
    inters : np.ndarray
        (3,) Intersection location. (np.inf, np.inf, np.inf) if no intersection.
    """
    ndotu = plane_n.dot(ray_dir)

    if abs(ndotu) < eps:
        return np.array([np.inf, np.inf, np.inf])
    w = ray_origin - plane_p
    si = -plane_n.dot(w) / ndotu
    intersec = w + si * ray_dir + plane_p

    return intersec
